"""Test for the Tektronix FCA3000 driver."""
import struct

from math import isnan

from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import call

from typing import cast

from dataclasses import dataclass


from qmi.core.transport import QMI_Transport
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.instruments.tektronix.fca3000 import Tektronix_FCA3000
from qmi.core.exceptions import QMI_InstrumentException


@dataclass
class TestMeta:
    """Test meta data."""

    transport: MagicMock
    transport_str: MagicMock
    name: MagicMock
    scpi: MagicMock
    super: MagicMock
    instr: Tektronix_FCA3000


class TestFCA3000(TestCase):
    """Testcase for the Tektronix_FCA3000 class."""

    def setUp(self):
        mock_transport = MagicMock(spec=QMI_Transport)
        mock_name = MagicMock()
        mock_transport_str = MagicMock()
        mock_scpi = MagicMock(spec=ScpiProtocol)
        mock_super = MagicMock()

        with patch(
            "qmi.instruments.tektronix.fca3000.create_transport",
            return_value=mock_transport,
        ), patch("qmi.instruments.tektronix.fca3000.ScpiProtocol", mock_scpi):
            instr = Tektronix_FCA3000(MagicMock(), mock_name, mock_transport_str)

        self._patcher_super = patch(
            "qmi.instruments.tektronix.fca3000.super", mock_super
        )
        self._patcher_super.start()

        self._meta = TestMeta(
            transport=mock_transport,
            instr=cast(Tektronix_FCA3000, instr),
            transport_str=mock_transport_str,
            name=mock_name,
            scpi=mock_scpi(mock_transport),
            super=mock_super,
        )

    def tearDown(self):
        self._meta = None
        self._patcher_super.stop()

    def test_init(self):
        """Tektronix_FCA3000.__init__(), happy flow."""
        self.assertEqual(self._meta.instr._transport, self._meta.transport)
        self.assertEqual(self._meta.instr._scpi_transport, self._meta.scpi)

    def test_open(self):
        """Tektronix_FCA3000.open(), happy flow."""
        self._meta.instr.open()
        self._meta.transport.open.assert_called_once_with()
        self._meta.super().open.assert_called_once_with()

    def test_close(self):
        """Tektronix_FCA3000.close(), happy flow."""
        self._meta.instr.close()
        self._meta.super().close.assert_called_once_with()
        self._meta.transport.close.assert_called_once_with()

    def test_reset(self):
        """Tektronix_FCA3000.reset(), happy flow."""
        self._meta.instr.reset()
        self._meta.scpi.write.assert_called_once_with("*RST")
        self._meta.scpi.ask.assert_called_once_with("*OPC?")

    def test_get_idn(self):
        """Tektronix_FCA3000.get_idn(), happy flow."""
        self._meta.scpi.ask = MagicMock(return_value="VENDOR,MODEL,SERIAL,VERSION")
        rt_val = self._meta.instr.get_idn()
        self._meta.scpi.ask.assert_called_once_with("*IDN?")
        self.assertEqual(rt_val.vendor, "VENDOR")
        self.assertEqual(rt_val.model, "MODEL")
        self.assertEqual(rt_val.serial, "SERIAL")
        self.assertEqual(rt_val.version, "VERSION")

    def test_get_idn_invalid_response(self):
        """Tektronix_FCA3000.get_idn(), invalid response."""
        self._meta.scpi.ask = MagicMock(return_value="")
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr.get_idn()
        self._meta.scpi.ask.assert_called_once_with("*IDN?")

    def test_get_errors(self):
        """Tektronix_FCA3000.get_errors(), happy flow."""
        self._meta.scpi.ask = MagicMock(side_effect=["1,", "2,", "0,"])
        rt_val = self._meta.instr.get_errors()
        self._meta.scpi.ask.assert_has_calls(
            [call("SYST:ERR?"), call("SYST:ERR?"), call("SYST:ERR?")]
        )
        self.assertEqual(len(rt_val), 2)
        self.assertEqual(rt_val[0], "1,")
        self.assertEqual(rt_val[1], "2,")

    def test_measure_frequency(self):
        """Tektronix_FCA3000.measure_frequency(), happy flow."""
        mock_channel = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value="0.003    ")
        rt_val = self._meta.instr.measure_frequency(mock_channel)
        self._meta.scpi.ask.assert_called_once_with(
            "MEAS:FREQ? (@{})".format(mock_channel)
        )
        self.assertEqual(rt_val, 0.003)

    def test_configure_frequency(self):
        """Tektronix_FCA3000.configure_frequency(), happy flow."""
        mock_channel = MagicMock()
        mock_aperture = 0.01
        mock_trigger_level = 0.02
        mock_timestamp = MagicMock()
        self._meta.instr.configure_frequency(
            channel=mock_channel,
            aperture=mock_aperture,
            trigger_level=mock_trigger_level,
            timestamp=mock_timestamp,
        )
        self._meta.scpi.write.assert_has_calls(
            [
                call("CONF:FREQ (@{})".format(mock_channel)),
                call("ACQ:APER {:.6g}".format(mock_aperture)),
                call("INP{}:LEV:AUTO 0".format(mock_channel)),
                call("INP{}:LEV {:.6g}".format(mock_channel, mock_trigger_level)),
                call("FORM:TINF {}".format(int(mock_timestamp))),
            ]
        )

    def test_read_values(self):
        """Tektronix_FCA3000.read_value(), happy flow."""
        self._meta.scpi.ask = MagicMock(return_value="0.001,0.002")
        rt_val = self._meta.instr.read_value()
        self._meta.scpi.ask.assert_called_once_with("READ?")
        self.assertEqual(rt_val, 0.001)

    def test_read_timestamped_value(self):
        """Tektronix_FCA3000.read_timestamped_value(), happy flow."""
        self._meta.scpi.ask = MagicMock(return_value="0.001,0.002")
        rt_val = self._meta.instr.read_timestamped_value()
        self._meta.scpi.ask.assert_called_once_with("READ?")
        self.assertEqual(rt_val[0], 0.002)
        self.assertEqual(rt_val[1], 0.001)

    def test_read_timestamped_value_invalid_response(self):
        """Tektronix_FCA3000.read_timestamped_value(), happy flow."""
        self._meta.scpi.ask = MagicMock(return_value="")
        with self.assertRaises(QMI_InstrumentException):
            self._meta.instr.read_timestamped_value()
        self._meta.scpi.ask.assert_called_once_with("READ?")

    def test_set_talk_only_enable(self):
        """Tektronix_FCA3000.set_talk_only(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        self._meta.instr.set_talk_only(True)
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.transport.read_until_timeout.assert_called_once_with(
            nbytes=16384, timeout=1.0
        )
        self._meta.scpi.ask.assert_called_once_with("*OPC?")
        self._meta.scpi.write.assert_has_calls(
            [
                call("++ifc"),
                call("FORM PACK"),
                call("DISP:ENAB 0"),
                call("INIT:CONT 1"),
                call("SYST:TALK 1"),
            ]
        )
        self._meta.scpi.read_binary_data.assert_called_once_with()

    def test_set_talk_only_disable(self):
        """Tektronix_FCA3000.set_talk_only(), happy flow."""
        self._meta.instr._check_is_open = MagicMock()
        self._meta.instr.set_talk_only(False)
        self._meta.instr._check_is_open.assert_called_once_with()
        self._meta.transport.read_until_timeout.assert_called_once_with(
            nbytes=16384, timeout=1.0
        )
        self._meta.scpi.ask.assert_called_once_with("*OPC?")
        self._meta.scpi.write.assert_called_once_with("++ifc")

    def test_get_value_talk_only(self):
        """Tektronix_FCA3000.get_value_talk_only(), happy flow."""
        mock_timestamp = 100
        mock_value = 200
        self._meta.scpi.read_binary_data = MagicMock(
            return_value=struct.pack(">dq", mock_value, mock_timestamp)
        )
        rt_val = self._meta.instr.get_value_talk_only()
        self.assertEqual(
            rt_val[0], mock_timestamp * 1.0e-12
        )  # timestamp conversion in function
        self.assertEqual(rt_val[1], mock_value)

    def test_get_value_talk_only_no_timestamp(self):
        """Tektronix_FCA3000.get_value_talk_only(), happy flow."""
        mock_value = 200
        self._meta.scpi.read_binary_data = MagicMock(
            return_value=struct.pack(">d", mock_value)
        )
        rt_val = self._meta.instr.get_value_talk_only()
        self.assertTrue(isnan(rt_val[0]))
        self.assertEqual(rt_val[1], mock_value)

    def test_get_trigger_level(self):
        """Tektronix_FCA3000.get_trigger_level(), happy flow."""
        mock_channel = MagicMock()
        self._meta.scpi.ask = MagicMock(return_value="0.001")
        rt_val = self._meta.instr.get_trigger_level(mock_channel)
        self._meta.scpi.ask.assert_called_once_with("INP{}:LEV?".format(mock_channel))
        self.assertEqual(rt_val, 0.001)

    def test_set_display_enabled(self):
        """Tektronix_FCA3000.set_display_enabled(), happy flow."""
        mock_enable = MagicMock()
        self._meta.instr.set_display_enabled(mock_enable)
        self._meta.scpi.write.assert_called_once_with(
            "DISP:ENAB {}".format(int(mock_enable))
        )

    def test_set_initiate_continuous(self):
        """Tektronix_FCA3000.set_initiate_continuous(), happy flow."""
        mock_enable = MagicMock()
        self._meta.instr.set_initiate_continuous(mock_enable)
        self._meta.scpi.write.assert_called_once_with(
            "INIT:CONT {}".format(int(mock_enable))
        )



"""
Digilent, USB oscilloscope and logic analyzer device.

The qmi.instruments.digilent package provides support for:
- Analog Discovery 2 device.
"""
from qmi.instruments.digilent.analog_discovery import OnClose, Filter
from qmi.instruments.digilent.analog_discovery import AnalogDiscovery2  # cannot remove due to legacy --> deprecate?
# Alternative, QMI naming convention approved names
Digilent_AnalogDiscovery2 = AnalogDiscovery2



"""Data structures for measurement data."""

import collections
import re
import time
from typing import Dict, List, Match, Optional, TextIO, Tuple, Union

import numpy as np
import h5py


class DataSet:
    """A dataset is a series of values obtained during a measurement.

    A dataset contains an array of values in the form of a N-dimensional Numpy array (N >= 2).

    The first (N-1) axes of the Numpy array represent independent variables or iterations in the measurement.
    Each of these axes may have an optional label, a physical unit, and a mapping of array indices to values
    on the physical axis.

    The last axis of the Numpy array acts as a "column index". Each column may have an associated label and
    physical unit.

    A dataset may have attributes. Each attribute has a name, which is a short string, unique to the dataset.
    Each attribute has a value which may be a string or a number.

    Reading or changing values in the dataset is done by directly accessing the Numpy array in the DataSet
    instance. For example:

        dataset.data[2, 0:5] += 1

    The following fields exist inside a DataSet instance. Application code may read or modify the contents of
    these fields directly. However the shape and data type of these fields must not be changed.

    Internal Variables:
        ~DataSet.name:         Name of the dataset.
        ~DataSet.data:         Numpy array containing the actual data.
        ~DataSet.timestamp:    POSIX time stamp associated with the data.
        axis_label:   List of strings specifying labels for the first (N-1) axes.
        axis_unit:    List of strings specifying units for the first (N-1) axes.
        axis_scale:   List of optional 1D Numpy arrays specifying value mappings for the first (N-1) axes.
        column_label: List of strings specifying column labels.
        column_unit:  List of strings specifying column units.
        attrs:        Dictionary of application-specific attributes.

    The entire dataset is kept in memory (RAM). This makes the dataset class unsuitable for very large amounts of data.
    """

    def __init__(self,
                 name: str,
                 shape: Optional[Tuple[int, ...]] = None,
                 dtype: Optional[Union[np.dtype, type]] = None,
                 data: Optional[np.ndarray] = None
                 ) -> None:
        """Initialize a new dataset.

        Parameters:
            name: Name of the dataset. This should be a short string without spaces or strange symbols,
                suitable for use as part of a file name.
            shape: Tuple of axis dimensions. Used to create a zero-initialized dataset if the actual data
                are not yet available. The last axis dimension represents the number of columns in the dataset.
            dtype: Type of value in each data point. If not specified, the default is np.float64.
            data: Optional Numpy array containing the actual data. The new dataset instance will contain a reference
                to the specified Numpy array. Modifying the Numpy array will cause the contents of the dataset to be
                changed as well.
        """

        self.name = name
        self.timestamp = time.time()

        if data is not None:
            # Check that the specified data is a Numpy array.
            if not isinstance(data, np.ndarray):
                raise TypeError("Specified 'data' parameter must be a Numpy array")

            # Check shape and data type.
            if shape is not None:
                if data.shape != tuple(shape):
                    raise ValueError("Data does not match specified shape")

            if dtype is not None:
                if data.dtype != dtype:
                    raise ValueError("Data does not match specified data type")

            # Copy array reference.
            self.data = data

        else:

            if shape is None:
                raise TypeError("Either 'shape' or 'data' parameter must be specified")

            if dtype is None:
                dtype = np.float64

            # Create zero-initialized array.
            self.data = np.zeros(tuple(shape), dtype=dtype)

        # Check shape.
        ndim = len(self.data.shape)
        if ndim < 2:
            raise ValueError("Dataset must have at least 2 axes")
        if np.min(self.data.shape) < 1:
            raise ValueError("Zero-size or negative size axes are not allowed")

        # Initialize axis labels.
        self.axis_label = [""]*(ndim - 1)
        self.axis_unit = [""]*(ndim - 1)
        self.axis_scale = [None]*(ndim - 1)  # type: List[Optional[np.ndarray]]

        # Initialize column labels.
        ncol = self.data.shape[-1]
        self.column_label = ncol * [""]
        self.column_unit = ncol * [""]

        # Initialize empty set of attributes.
        self.attrs = {}  # type: Dict[str, Union[str, int, float]]

    def set_axis_label(self, axis: int, label: str) -> None:
        """Specify an axis label.

        Parameters:
            axis: int - axis number (0, 1, ...)
            label: str - label string of the axis
        """
        if not isinstance(axis, int):
            raise TypeError("Parameter 'axis' must be an integer")

        if axis < 0 or axis >= len(self.axis_label):
            raise ValueError("Invalid value for parameter 'axis'")

        self.axis_label[axis] = label

    def set_axis_unit(self, axis: int, unit: str) -> None:
        """Specify the physical unit for an axis.

        Parameters:
            axis: int - axis number (0, 1, ...)
            unit: str - unit string of the axis
        """
        if not isinstance(axis, int):
            raise TypeError("Parameter 'axis' must be an integer")

        if axis < 0 or axis >= len(self.axis_unit):
            raise ValueError("Invalid value for parameter 'axis'")

        self.axis_unit[axis] = unit

    def set_axis_scale(self, axis: int, scale: np.ndarray) -> None:
        """Specify a mapping from array indices to physical values along an axis.

        Parameters:
            axis: Axis to which the mapping applies (the first axis has number 0).
            scale: 1D Numpy array of values along the axis. The length must match the size of the axis.
        """
        if not isinstance(axis, int):
            raise TypeError("Parameter 'axis' must be an integer")

        if axis < 0 or axis >= len(self.data.shape) - 1:
            raise ValueError("Invalid value for parameter 'axis'")

        v = np.array(scale)
        if v.shape != (self.data.shape[axis],):
            raise ValueError("Invalid shape for scale array")

        if not np.all(np.isfinite(scale)):
            raise ValueError("Only finite values allowed on the axis scale")

        self.axis_scale[axis] = scale

    def set_column_label(self, col: int, label: str) -> None:
        """Specify a label for a column in a multi-column data set.

        Parameters:
            col: int - column number (0, 1, ...)
            label: str - column label string
        """
        if not isinstance(col, int):
            raise TypeError("Parameter 'col' must be an integer")
        if col < 0 or col >= len(self.column_label):
            raise ValueError("Invalid value for parameter 'col'")
        self.column_label[col] = label

    def set_column_unit(self, col: int, unit: str) -> None:
        """Specify a physical unit for a column in a multi-column data set.

        Parameters:
            col: int - column number (0, 1, ...)
            unit: str - column unit string
        """
        if not isinstance(col, int):
            raise TypeError("Parameter 'col' must be an integer")

        if col < 0 or col >= len(self.column_unit):
            raise ValueError("Invalid value for parameter 'col'")

        self.column_unit[col] = unit


def _parse_attribute_value(s: str) -> Union[int, float, str]:
    """Parse an attribute value.

    This function should be able to evaluate any string
    produced by repr() when acting on a string, int or float.

    Parameters:
        s: str - the string to parse
    """

    def replace_esc(m: Match[str]) -> str:
        replace_table = {
            "\n": "",
            "\\": "\\",
            "'": "'",
            '"': '"',
            "a": "\a",
            "b": "\b",
            "f": "\f",
            "n": "\n",
            "r": "\r",
            "t": "\t",
            "v": "\v"
        }
        t = m.group(1)
        if t in replace_table:
            return replace_table[t]
        elif t.startswith("x") or t.startswith("u"):
            return chr(int(t[1:], 16))
        else:
            return chr(int(t, 8))

    if ("'" in s) or ('"' in s):
        # Quoted string.
        quote_char = s[0]
        if quote_char not in ('"', "'"):
            raise ValueError("Invalid attribute value syntax {!r}".format(s))
        if not s[1:].endswith(quote_char):
            raise ValueError("Invalid attribute value syntax {!r}".format(s))
        # Strip quotes.
        s = s[1:-1]
        # Check for non-escaped occurrences of quote symbol.
        t = re.sub("\\\\.", "", s)
        if quote_char in t:
            raise ValueError("Invalid attribute value syntax {!r}".format(s))
        # Expand escape sequences.
        s = re.sub("\\\\(['\"abfnrtv]|\\\\|[0-7]{1,3}|x[0-9a-fA-F]{2}|u[0-9a-fA-F]{4})", replace_esc, s)
        return s

    elif re.match(r"^[+-]?[0-9]*$", s):
        # Integer literal.
        return int(s)

    elif (s == 'True') or (s == 'False'):
        # Must be a boolean, one line return bool value.
        return s == 'True'

    else:
        # Must be floating point literal.
        return float(s)


def write_dataset_to_hdf5(dataset: DataSet, hdf_group: h5py.Group) -> None:
    """Write the specified dataset to the specified HDF5 group.

    The dataset "name" field determines the name of the corresponding HDF5 dataset.
    An error occurs if the HDF5 group already contains a dataset with the same name.

    Note that this function may create additional supporting datasets in the HDF5 group if the DataSet instance
    uses axis scales. In this case, HDF5 datasets named "<datasetname>_axisN_scale" will be created in
    addition to the main dataset.

    Parameters:
        dataset: DataSet instance to write to HDF5.
        hdf_group: HDF5 File or Group instance to which the dataset is written.
    """

    ndim = len(dataset.data.shape)
    ncol = dataset.data.shape[-1]

    ds = hdf_group.create_dataset(dataset.name, data=dataset.data)

    # Special timestamp attribute.
    ds.attrs["QMI_DataSet_timestamp"] = dataset.timestamp
    ds.attrs["QMI_DataSet_time_str"] = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(dataset.timestamp))

    # Special attributes for axis labels / units.
    for axis in range(ndim - 1):
        if dataset.axis_label[axis]:
            ds.attrs["QMI_DataSet_axis{}_label".format(axis)] = dataset.axis_label[axis]

        if dataset.axis_unit[axis]:
            ds.attrs["QMI_DataSet_axis{}_unit".format(axis)] = dataset.axis_unit[axis]

    # Special attributes for column labels / units.
    for col in range(ncol):
        if dataset.column_label[col]:
            ds.attrs["QMI_DataSet_column{}_label".format(col)] = dataset.column_label[col]

        if dataset.column_unit[col]:
            ds.attrs["QMI_DataSet_column{}_unit".format(col)] = dataset.column_unit[col]

    # Dimension scales.
    for axis in range(ndim - 1):
        if dataset.axis_label[axis]:
            ds.dims[axis].label = dataset.axis_label[axis]

        if dataset.axis_scale[axis] is not None:
            # Create an extra dataset to hold the dimension scale.
            scale_name = dataset.name + "_axis{}_scale".format(axis)
            ds_scale = hdf_group.create_dataset(scale_name, data=dataset.axis_scale[axis])
            # Attach the dimension scale to the axis.
            ds_scale.make_scale(scale_name)
            ds.dims[axis].attach_scale(ds_scale)

    # Custom attributes.
    for (name, value) in dataset.attrs.items():
        if name.startswith("QMI_DataSet") or name.startswith("DIMENSION_"):
            raise ValueError("Invalid use of special attribute name {!r}".format(name))

        ds.attrs[name] = value

    # Special attribute to recognize format.
    ds.attrs["QMI_DataSet"] = 1


def read_dataset_from_hdf5(ds: h5py.Dataset) -> DataSet:
    """Extract a DataSet instance from the specified HDF5 dataset.

    Note that this function may fetch additional HDF5 datasets from
    the parent HDF5 group if the dataset uses dimension scales.

    Parameters:
        ds: HDF5 Dataset instance to read from.

    Returns:
        DataSet instance.
    """

    # Check that the HDF5 dataset was created by this Python module.
    if ds.attrs.get("QMI_DataSet") != 1:
        raise ValueError("HDF5 dataset not in expected format")

    # Sanity check.
    if (len(ds.shape) < 2) or np.min(ds.shape) < 1:
        raise ValueError("Invalid shape of HDF5 dataset")

    # Create DataSet instance and read actual data.
    name = ds.name.split("/")[-1]
    dataset = DataSet(name=name, data=ds[:])

    ndim = len(dataset.data.shape)
    ncol = dataset.data.shape[-1]

    # Read timestamp.
    dataset.timestamp = ds.attrs["QMI_DataSet_timestamp"]

    # Read special attributes for labels.
    for axis in range(ndim - 1):
        dataset.axis_label[axis] = ds.attrs.get("QMI_DataSet_axis{}_label".format(axis), "")
        dataset.axis_unit[axis] = ds.attrs.get("QMI_DataSet_axis{}_unit".format(axis), "")

    for col in range(ncol):
        dataset.column_label[col] = ds.attrs.get("QMI_DataSet_column{}_label".format(col), "")
        dataset.column_unit[col] = ds.attrs.get("QMI_DataSet_column{}_unit".format(col), "")

    # Read dimension scales.
    for axis in range(ndim - 1):
        if len(ds.dims[axis]) > 0:
            scale = ds.dims[axis][0]
            if scale.shape != (dataset.data.shape[axis],):
                raise ValueError("Invalid shape of dimension scale for axis {}".format(axis))
            dataset.axis_scale[axis] = scale[:]

    # Read custom attributes.
    for (name, value) in ds.attrs.items():
        if (not name.startswith("QMI_DataSet")) and (not name.startswith("DIMENSION_")):
            dataset.attrs[name] = value

    return dataset


def write_dataset_to_text(dataset: DataSet, fh: TextIO) -> None:
    """Write the specified dataset to a text file.

    Note that this function may create additional supporting datasets in
    the HDF5 group if the DataSet instance uses axis scales. In this case,
    HDF5 datasets named "<datasetname>_axisN_scale" will be created in
    addition to the main dataset.

    Parameters:
        dataset: DataSet instance to write to HDF5.
        fh: File handle open for writing in text mode.
    """

    ndim = len(dataset.data.shape)
    ncol = dataset.data.shape[-1]

    # Create special columns if needed.
    special_column_label = []
    special_column_unit = []

    if ndim > 2:
        # Create special axis index columns.
        for axis in range(ndim - 1):
            special_column_label.append("axis{}_index".format(axis))
            special_column_unit.append("")

    # Create special axis scale columns if needed.
    for axis in range(ndim - 1):
        if dataset.axis_scale[axis] is not None:
            special_column_label.append("axis{}_scale".format(axis))
            special_column_unit.append(dataset.axis_unit[axis])

    # Prepare attributes.
    attrs = collections.OrderedDict()  # type: Dict[str, Union[int, float, str]]

    # Dataset name.
    attrs["QMI_DataSet_name"] = dataset.name

    # Timestamp.
    attrs["QMI_DataSet_timestamp"] = dataset.timestamp
    attrs["QMI_DataSet_time_str"] = time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(dataset.timestamp))

    # Data shape.
    attrs["QMI_DataSet_ndim"] = ndim
    attrs["QMI_DataSet_ncol"] = ncol

    # Axis labels / units.
    for axis in range(ndim - 1):
        attrs["QMI_DataSet_axis{}_size".format(axis)] = dataset.data.shape[axis]
        if dataset.axis_label[axis]:
            attrs["QMI_DataSet_axis{}_label".format(axis)] = dataset.axis_label[axis]
        if dataset.axis_unit[axis]:
            attrs["QMI_DataSet_axis{}_unit".format(axis)] = dataset.axis_unit[axis]

    # Column labels / units.
    column_label = special_column_label + dataset.column_label
    column_unit = special_column_unit + dataset.column_unit
    for col in range(len(column_label)):
        if column_label[col]:
            attrs["QMI_DataSet_column{}_label".format(col)] = column_label[col]

        if column_unit[col]:
            attrs["QMI_DataSet_column{}_unit".format(col)] = column_unit[col]

    # Custom attributes.
    for (name, val) in dataset.attrs.items():
        if name.startswith("QMI_DataSet"):
            raise ValueError("Invalid use of special attribute name {!r}".format(name))

        if not name:
            raise ValueError("Invalid use of empty attribute name")

        if ':' in name:
            raise ValueError("Invalid character ':' in attribute name {!r}".format(name))

        attrs[name] = val

    # Reshape data to 2D format.
    if ndim > 2:
        nrow = np.prod(dataset.data.shape[:-1])
        rawdata = dataset.data.reshape((nrow, ncol))

    else:
        rawdata = dataset.data

    # Insert axis index columns.
    extra_columns = []
    if ndim > 2:
        for axis in range(ndim - 1):
            n = dataset.data.shape[axis]
            outer_rows = int(np.prod(dataset.data.shape[:axis], dtype=np.int32))
            inner_rows = int(np.prod(dataset.data.shape[axis+1:-1], dtype=np.int32))
            extra_columns.append(np.tile(np.repeat(np.arange(n), inner_rows), outer_rows))

    # Insert axis scale columns.
    for axis in range(ndim - 1):
        dataset_axis_scale = dataset.axis_scale[axis]
        if dataset_axis_scale is not None:
            assert dataset_axis_scale.shape == (dataset.data.shape[axis],)
            outer_rows = int(np.prod(dataset.data.shape[:axis], dtype=np.int32))
            inner_rows = int(np.prod(dataset.data.shape[axis+1:-1], dtype=np.int32))
            extra_columns.append(np.tile(np.repeat(dataset_axis_scale, inner_rows), outer_rows))

    if extra_columns:
        rawdata = np.column_stack(extra_columns + [rawdata])

    # Write marker line.
    fh.write("# QMI_DataSet\n")
    fh.write("#\n")

    # Write attributes.
    for (name, value) in attrs.items():
        fh.write("# {}: {!r}\n". format(name, value))

    fh.write("#\n")

    # Write actual data.
    np.savetxt(fh, rawdata)


def read_dataset_from_text(fh: TextIO) -> DataSet:
    """Read a DataSet instance from a text file.

    Parameters:
        fh: File handle open for reading in text mode.

    Returns:
        DataSet instance.
    """

    # Check marker line.
    line = fh.readline().strip()
    if line != "# QMI_DataSet":
        raise ValueError("Invalid file format; expecting marker but got {!r}".format(line))

    line = fh.readline().strip()
    if line != "#":
        raise ValueError("Invalid file format; expecting separator but got {!r}".format(line))

    # Read attributes.
    attrs = {}
    while True:
        line = fh.readline().strip()
        # Stop at separator between attributes and data.
        if line == "#":
            break
        # Read attribute.
        p = line.find(":")
        if (not line.startswith("# ")) or (p < 0):
            raise ValueError("Invalid file format; expecting attribute but got {!r}".format(line))

        name = line[2:p]
        value = line[p+1:].strip()
        if not name:
            raise ValueError("Invalid file format; expecting attribute but got {!r}".format(line))

        attrs[name] = _parse_attribute_value(value)

    # Read raw data.
    rawdata = np.loadtxt(fh, ndmin=2)
    (nrow, total_columns) = rawdata.shape

    # Determine dataset name.
    dataset_name = attrs.get("QMI_DataSet_name")
    if not isinstance(dataset_name, str):
        raise ValueError("Missing required attribute QMI_DataSet_name")

    # Determine dataset shape.
    ndim = int(attrs["QMI_DataSet_ndim"])
    ncol = int(attrs["QMI_DataSet_ncol"])
    assert ndim >= 2

    shape_list: List[int] = []
    for axis in range(ndim - 1):
        axis_size = attrs["QMI_DataSet_axis{}_size".format(axis)]
        if not isinstance(axis_size, int):
            raise ValueError("Invalid value for attribute QMI_DataSet_axis{}_size".format(axis))
        shape_list.append(axis_size)
    shape_list.append(ncol)
    shape = tuple(shape_list)

    # Verify number of rows.
    expect_rows = np.prod(shape[:-1])
    if nrow != expect_rows:
        raise ValueError("Expecting {} rows but got {} rows".format(expect_rows, nrow))

    # Verify number of columns.
    if total_columns < ncol:
        raise ValueError("Expecting at least {} columns but got {} columns".format(ncol, total_columns))
    num_special_columns = total_columns - ncol

    # Extract and reshape actual data.
    data = rawdata[:, num_special_columns:].reshape(*shape)

    # Create DataSet instance.
    dataset = DataSet(name=dataset_name, data=data)

    # Set timestamp.
    dataset.timestamp = float(attrs["QMI_DataSet_timestamp"])

    # Set axis labels and units.
    for axis in range(ndim - 1):
        dataset.axis_label[axis] = str(attrs.get("QMI_DataSet_axis{}_label".format(axis), ""))
        dataset.axis_unit[axis] = str(attrs.get("QMI_DataSet_axis{}_unit".format(axis), ""))

    for col in range(ncol):
        dataset.column_label[col] = str(attrs.get("QMI_DataSet_column{}_label".format(num_special_columns + col), ""))
        dataset.column_unit[col] = str(attrs.get("QMI_DataSet_column{}_unit".format(num_special_columns + col), ""))

    # Verify index columns and extract axis scales.
    for col in range(num_special_columns):
        label = attrs.get("QMI_DataSet_column{}_label".format(col))
        if not isinstance(label, str):
            raise ValueError("Missing label for special column {}".format(col))

        if label.startswith("axis") and label.endswith("_index"):
            # Verify index column.
            axis = int(label[4:-6])
            n = dataset.data.shape[axis]
            outer_rows = int(np.prod(dataset.data.shape[:axis], dtype=np.int32))
            inner_rows = int(np.prod(dataset.data.shape[axis+1:-1], dtype=np.int32))
            idx = np.tile(np.repeat(np.arange(n), inner_rows), outer_rows)
            if not np.all(rawdata[:, col].astype(np.int64) == idx):
                raise ValueError("Inconsistent index data for axis {}".format(axis))

        elif label.startswith("axis") and label.endswith("_scale"):
            # Set axis scale.
            axis = int(label[4:-6])
            n = dataset.data.shape[axis]
            outer_rows = int(np.prod(dataset.data.shape[:axis], dtype=np.int32))
            inner_rows = int(np.prod(dataset.data.shape[axis+1:-1], dtype=np.int32))
            scale = rawdata[0:n*inner_rows:inner_rows, col]
            scale_raw = np.tile(np.repeat(scale, inner_rows), outer_rows)
            if not np.all(rawdata[:, col] == scale_raw):
                raise ValueError("Inconsistent scale data for axis {}".format(axis))

            dataset.axis_scale[axis] = scale

    # Read custom attributes.
    for attribute_name, attribute_value in attrs.items():
        if not attribute_name.startswith("QMI_DataSet"):
            dataset.attrs[attribute_name] = attribute_value

    return dataset



#!/usr/bin/env python3

"""Unit tests for qmi.core.config_struct."""

import unittest
from dataclasses import field
from typing import Any, Dict, List, Optional, Tuple, Union

from qmi.core.config_struct import configstruct, config_struct_from_dict, config_struct_to_dict
from qmi.core.exceptions import QMI_ConfigurationException


class TestConfigStruct(unittest.TestCase):

    def test_01_parse_simple(self):
        """Test parsing a simple structure."""

        @configstruct
        class MyConfig:
            i:  int
            s:  str
            f:  float
            b:  bool

        data1 = {
            "i": 42,
            "s": "hello",
            "f": 3.1415,
            "b": False
        }

        data2 = {
            "i": -1,
            "s": "",
            "f": 1.0,
            "b": True
        }

        cfg1 = config_struct_from_dict(data1, MyConfig)
        cfg2 = config_struct_from_dict(data2, MyConfig)

        self.assertIsInstance(cfg1, MyConfig)
        self.assertIsInstance(cfg1.i, int)
        self.assertIsInstance(cfg1.s, str)
        self.assertIsInstance(cfg1.f, float)
        self.assertIsInstance(cfg1.b, bool)
        self.assertEqual(cfg1.i, 42)
        self.assertEqual(cfg1.s, "hello")
        self.assertEqual(cfg1.f, 3.1415)
        self.assertEqual(cfg1.b, False)

        self.assertIsInstance(cfg2, MyConfig)
        self.assertIsInstance(cfg2.i, int)
        self.assertIsInstance(cfg2.s, str)
        self.assertIsInstance(cfg2.f, float)
        self.assertIsInstance(cfg2.b, bool)
        self.assertEqual(cfg2.i, -1)
        self.assertEqual(cfg2.s, "")
        self.assertEqual(cfg2.f, 1.0)
        self.assertEqual(cfg2.b, True)

    def test_02_serialize_simple(self):
        """Test serializing a simple structure."""

        @configstruct
        class MyConfig:
            i:  int
            s:  str
            f:  float
            b:  bool

        cfg1 = MyConfig(i=11, s="sss", f=2.71, b=True)
        data1 = {"i": 11, "s": "sss", "f": 2.71, "b": True}

        q1 = config_struct_to_dict(cfg1)
        self.assertIsInstance(data1, dict)
        self.assertEqual(q1, data1)

    def test_03_float_conversion(self):
        """Test automatic promotion from int to float."""

        @configstruct
        class MyConfig:
            i:  int
            f:  float

        data1 = {"i": 5, "f": 3.0}
        data2 = {"i": 10, "f": -8}

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertIsInstance(cfg1.f, float)
        self.assertEqual(cfg1, MyConfig(i=5, f=3.0))

        cfg2 = config_struct_from_dict(data2, MyConfig)
        self.assertIsInstance(cfg2.f, float)
        self.assertEqual(cfg2, MyConfig(i=10, f=-8.0))

        q1 = config_struct_to_dict(cfg1)
        self.assertEqual(q1, data1)

        q2 = config_struct_to_dict(cfg2)
        self.assertIsInstance(q2["f"], float)
        self.assertEqual(q2, data2)

    def test_04_class_not_dataclass(self):
        """The configuration class must be a dataclass"""
        class MyConfig:
            i:  int
            f:  float

        data1 = {"i": 5, "f": 3.0, "3": "str"}

        exp_error = "Configuration class type must be a dataclass"
        exp_error2 = "Configuration data must be a dataclass instance"
        with self.assertRaises(TypeError) as type_err:
            config_struct_from_dict(data1, MyConfig)

        with self.assertRaises(TypeError) as type_err2:
            config_struct_to_dict(MyConfig)

        self.assertEqual(exp_error, str(type_err.exception))
        self.assertEqual(exp_error2, str(type_err2.exception))

    def test_05_no_keyword_arguments_allowed(self):
        """The configstruct doesn't allow keyword arguments"""
        @configstruct
        class MyConfig:
            i:  int
            f:  float

        exp_error = "config_struct_from_dict() got an unexpected keyword argument 'kwarg'"
        data1 = {"i": 5, "f": 3.0}

        with self.assertRaises(TypeError) as kw_err:
            config_struct_from_dict(data1, MyConfig, kwarg=None)

        self.assertEqual(exp_error, str(kw_err.exception))

    def test_06_unsupported_config_type(self):
        """configstruct dictionary inputs allow only dictionaries with keys as strings."""
        @configstruct
        class MyConfig:
            d:  Dict[str, int]

        dict1 = {"d": {"0": 5, 1: 3.0}}
        exp_error = "Unsupported non-string dictionary key 1 in configuration item d"
        with self.assertRaises(QMI_ConfigurationException) as k_err:
            config_struct_from_dict(dict1, MyConfig)

        self.assertEqual(exp_error, str(k_err.exception))

    def test_07_missing_dataclass_input(self):
        """Call to get config_struct_from_dict errors when the respective class is 'forgotten' from inputs."""
        exp_error = "config_struct_from_dict() missing 1 required positional argument: 'cls'"
        with self.assertRaises(TypeError) as kw_err:
            config_struct_from_dict({})

        self.assertEqual(exp_error, str(kw_err.exception))

    def test_08_init_missing_argument(self):
        """The configstruct misses an argument"""
        class MyConfig:
            d:  Dict[str, int]

        exp_error = "MyConfig.__init__() missing required argument 'd'"
        cs = configstruct(MyConfig)
        with self.assertRaises(TypeError) as miss_err:
            cs()

        self.assertEqual(exp_error, str(miss_err.exception))

    def test_09_unsupported_value_type_inner(self):
        """Unsupported value type in _inner_config_struct_to_dict excepts."""
        @configstruct
        class MyConfig:
            i:  int

        exp_error = "Unsupported value type: b'5'"

        data1 = {"i": 5}
        cfg1 = config_struct_from_dict(data1, MyConfig)
        cfg1.i = b"5"  # Overwrite the value with an unsupported type
        with self.assertRaises(TypeError) as type_err:
            config_struct_to_dict(cfg1)

        self.assertEqual(exp_error, str(type_err.exception))

    def test_10_unsupported_data_type_in_config(self):
        """Unsupported data type in configstruct excepts."""
        @configstruct
        class MyConfig:
            t:  tuple

        exp_error = "Unsupported data type in configuration field t"

        data1 = {"t": [5,]}
        with self.assertRaises(QMI_ConfigurationException) as cfg_err:
            config_struct_from_dict(data1, MyConfig)

        self.assertEqual(exp_error, str(cfg_err.exception))

    def test_11_default_values(self):
        """Test a structure with default values for some fields."""

        @configstruct
        class MyConfig:
            a:  int
            b:  int = 111
            c:  str = "hello"
            d:  bool = True

        data1 = {"a": 1, "b": 2, "c": "howdie", "d": False}
        data2 = {"a": 5}
        data3 = {"a": 6, "c": "", "d": True}

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertEqual(cfg1.a, 1)
        self.assertEqual(cfg1.b, 2)
        self.assertEqual(cfg1.c, "howdie")
        self.assertEqual(cfg1.d, False)

        cfg2 = config_struct_from_dict(data2, MyConfig)
        self.assertEqual(cfg2.a, 5)
        self.assertEqual(cfg2.b, 111)
        self.assertEqual(cfg2.c, "hello")
        self.assertEqual(cfg2.d, True)

        cfg3 = config_struct_from_dict(data3, MyConfig)
        self.assertEqual(cfg3.a, 6)
        self.assertEqual(cfg3.b, 111)
        self.assertEqual(cfg3.c, "")
        self.assertEqual(cfg3.d, True)

        q1 = config_struct_to_dict(cfg1)
        self.assertEqual(q1, data1)

        q2 = config_struct_to_dict(cfg2)
        self.assertEqual(q2, {"a": 5, "b": 111, "c": "hello", "d": True})

        q3 = config_struct_to_dict(cfg3)
        self.assertEqual(q3, {"a": 6, "b": 111, "c": "", "d": True})

    def test_12_default_before_required(self):
        """Test a structure with default fields before required fields."""

        @configstruct
        class MyConfig:
            a:  int = 25
            b:  str

        data1 = {"a": 100, "b": "hello"}
        data2 = {"b": "bye"}

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertEqual(cfg1, MyConfig(a=100, b="hello"))

        cfg2 = config_struct_from_dict(data2, MyConfig)
        self.assertEqual(cfg2, MyConfig(a=25, b="bye"))

    def test_13_advanced_types(self):
        """Test a structure with non-scalar field types."""

        @configstruct
        class MyConfig:
            v:  list
            d:  dict
            vs: List[str]
            di: Dict[str, int]
            t:  Tuple[int, str]
            a:  Any
            opt: Optional[int]

        data1 = {
            "v": [101, 102, 103],
            "d": {"a": "aap", "n": "noot"},
            "vs": ["one", "two"],
            "di": {"one": 1, "thousand": 1000},
            "t": [5, "qqq"],
            "a": [15],
            "opt": 25
        }

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertEqual(cfg1, MyConfig(
            v=[101, 102, 103],
            d={"a": "aap", "n": "noot"},
            vs=["one", "two"],
            di={"one": 1, "thousand": 1000},
            t=(5, "qqq"),
            a=[15],
            opt=25))

        q1 = config_struct_to_dict(cfg1)

        self.assertEqual(q1, data1)

    def test_14_advanced_defaults(self):
        """Test a structure with default values for non-scalar fields."""

        @configstruct
        class MyConfig:
            v:  list            = field(default_factory=list)
            d:  dict            = field(default_factory=dict)
            vs: List[str]       = field(default_factory=list)
            di: Dict[str, int]  = field(default_factory=dict)
            t:  Tuple[int, str] = (0, "")
            a:  Any             = False
            opt: Optional[int]  = None

        data1 = {
            "v": [101, 102, 103],
            "d": {"a": "aap", "n": "noot"}
        }

        data2 = {
            "vs": ["one", "two"],
            "di": {"one": 1, "thousand": 1000},
            "t": [5, "qqq"],
            "a": [15],
            "opt": 25
        }

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertEqual(cfg1, MyConfig(
            v=[101, 102, 103],
            d={"a": "aap", "n": "noot"},
            vs=[],
            di={},
            t=(0, ""),
            a=False,
            opt=None))

        cfg2 = config_struct_from_dict(data2, MyConfig)
        self.assertEqual(cfg2, MyConfig(
            vs=["one", "two"],
            di={"one": 1, "thousand": 1000},
            t=(5, "qqq"),
            a=[15],
            opt=25))

        q1 = config_struct_to_dict(cfg1)
        self.assertEqual(q1, {
            "v": [101, 102, 103],
            "d": {"a": "aap", "n": "noot"},
            "vs": [],
            "di": {},
            "t": [0, ""],
            "a": False,
            "opt": None
        })

        q2 = config_struct_to_dict(cfg2)
        self.assertEqual(q2, {
            "v": [],
            "d": {},
            "vs": ["one", "two"],
            "di": {"one": 1, "thousand": 1000},
            "t": [5, "qqq"],
            "a": [15],
            "opt": 25
        })

    def test_15_nested_types(self):
        """Test a structure with default values for non-scalar fields."""

        @configstruct
        class MyConfig:
            vvi: List[List[int]]
            dvs: Dict[str, List[str]]
            optv: Optional[List[float]] = None

        data1 = {
            "vvi": [[], [1], [2, 3]],
            "dvs": {"one": ["1"], "two": ["00", "11"]},
            "optv": None
        }

        data2 = {
            "vvi": [[], [0, 0, 0], []],
            "dvs": {},
            "optv": [3.14, 1.4142]
        }

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertEqual(cfg1, MyConfig(
            vvi=[[], [1], [2, 3]],
            dvs={"one": ["1"], "two": ["00", "11"]},
            optv=None))

        cfg2 = config_struct_from_dict(data2, MyConfig)
        self.assertEqual(cfg2, MyConfig(
            vvi=[[], [0, 0, 0], []],
            dvs={},
            optv=[3.14, 1.4142]))

        q1 = config_struct_to_dict(cfg1)
        self.assertEqual(q1, data1)

        q2 = config_struct_to_dict(cfg2)
        self.assertEqual(q2, data2)

    def test_16_tuple_t_dotdotdot(self):
        """Test a Tuple[T, ...] structure field type."""

        @configstruct
        class MyConfig:
            t:  Tuple[int, ...]

        data1 = {
            "t": [5, 4, 3, 2, 1],
            }

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertEqual(cfg1, MyConfig(t=(5, 4, 3, 2, 1),))

        q1 = config_struct_to_dict(cfg1)

        self.assertEqual(q1, data1)

    def test_17_typing_types(self):
        """Test a structure with unspecified Typing field types."""

        @configstruct
        class MyConfig:
            vs: List
            di: Dict
            t: Tuple

        data1 = {
            "vs": ["one", "two"],
            "di": {"one": 1, "thousand": 1000},
            "t": (5, "qqq"),
        }

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertEqual(cfg1, MyConfig(
            vs=["one", "two"],
            di={"one": 1, "thousand": 1000},
            t=(5, "qqq"),
        ))

        # NOTE: When transforming back to dict, tuple(s) are turned to list(s) in _dictify_list_value.
        q1 = config_struct_to_dict(cfg1)
        data1.update({"t": [5, "qqq"]})
        self.assertEqual(q1, data1)

    def test_21_sub_structs(self):
        """Test a structure with sub-structures."""

        @configstruct
        class MySub:
            x: int
            y: float

        @configstruct
        class MyConfig:
            s: MySub
            v: List[MySub]

        data1 = {
            "s": {"x": 100, "y": 3.1415},
            "v": [
                {"x": 101, "y": 1.0},
                {"x": 102, "y": -1.0e6}
            ]
        }

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertIsInstance(cfg1, MyConfig)
        self.assertIsInstance(cfg1.s, MySub)
        self.assertIsInstance(cfg1.v[0], MySub)
        self.assertEqual(cfg1.s, MySub(x=100, y=3.1415))
        self.assertEqual(cfg1.v, [MySub(x=101, y=1.0),
                                  MySub(x=102, y=-1.0e6)])

    def test_22_extended_structs(self):
        """Test extended structure types."""

        @configstruct
        class MySubBase:
            x: int

        @configstruct
        class MySubExt(MySubBase):
            y: float

        @configstruct
        class MyBase:
            i: int
            s: str
            d: MySubBase

        @configstruct
        class MyConfig(MyBase):
            f: float
            s: str = "default"
            d: MySubExt

        data1 = {
            "i": 21,
            "f": 1.234,
            "s": "hello",
            "d": {"x": 10, "y": 3.14}
        }

        data2 = {
            "i": 50,
            "f": 123.4,
            "d": {"x": 11, "y": 1.111}
        }

        cfg1 = config_struct_from_dict(data1, MyConfig)
        self.assertEqual(cfg1, MyConfig(i=21, f=1.234, s="hello", d=MySubExt(x=10, y=3.14)))

        cfg2 = config_struct_from_dict(data2, MyConfig)
        self.assertEqual(cfg2, MyConfig(i=50, f=123.4, s="default", d=MySubExt(x=11, y=1.111)))

        q1 = config_struct_to_dict(cfg1)
        self.assertEqual(q1, data1)

        q2 = config_struct_to_dict(cfg2)
        self.assertEqual(q2, {
            "i": 50,
            "f": 123.4,
            "s": "default",
            "d": {"x": 11, "y": 1.111}
        })

    def test_31_fail_missing_field(self):
        """Fail on missing field."""

        @configstruct
        class MySub:
            p: int

        @configstruct
        class MyConfig:
            x: int
            y: Optional[float]  # NOTE : this field is not optional because it has no default
            d: MySub

        data1 = {}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data1, MyConfig)

        data2 = {"x": 25, "d": {"p": 1}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data2, MyConfig)

        data3 = {"y": 1.0, "d": {"p": 1}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data3, MyConfig)

        data4 = {"x": 25, "y": 1.0}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data4, MyConfig)

        data5 = {"x": 25, "y": 1.0, "d": {}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data5, MyConfig)

    def test_32_fail_unknown_field(self):
        """Fail on unknown field name."""

        @configstruct
        class MySub:
            p: int

        @configstruct
        class MyConfig:
            x: int
            d: MySub

        data1 = {"x": 21, "d": {"p": 100}, "yy": 22}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data1, MyConfig)

        data2 = {"x": 21, "d": {"p": 100, "qq": 101}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data2, MyConfig)

    def test_33_fail_bad_type(self):
        """Fail on incorrect value for field type."""

        @configstruct
        class MySub:
            b: bool

        @configstruct
        class MyConfig:
            v: List[int]
            s: str
            d: MySub
            m: Dict[str, int] = field(default_factory=dict)

        data1 = {"v": 5, "s": "hello", "d": {"b": False}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data1, MyConfig)

        data2 = {"v": [5], "s": 0, "d": {"b": False}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data2, MyConfig)

        data3 = {"v": [5], "s": "hello", "d": {"b": 0}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data3, MyConfig)

        data4 = {"v": [5, "6"], "s": "hello", "d": {"b": False}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data4, MyConfig)

        data5 = {"v": [], "s": "hello", "d": {"b": False}, "m": {"a": "b"}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data5, MyConfig)

    def test_34_fail_unsupported_type(self):
        """Fail on unsupported field type in struct definition."""

        @configstruct
        class MyConfig1:
            x: Union[list, tuple]  # union not supported

        data1 = {"x": []}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data1, MyConfig1)

        @configstruct
        class MyConfig2:
            x: Dict[int, int]  # integer dict key not supported

        data2 = {"x": {}}
        with self.assertRaises(QMI_ConfigurationException):
            config_struct_from_dict(data2, MyConfig2)

    def test_41_direct_instantiation_simple(self):
        """Test direct instantiation of a configstruct object."""

        @configstruct
        class MyConfig:
            x: int
            y: str

        x_val = 5
        y_val = "hello"
        my_config = MyConfig(x=x_val, y=y_val)

        self.assertEqual(my_config.x, x_val)
        self.assertEqual(my_config.y, y_val)

    def test_42_direct_instantiation_complex(self):
        """Test direct instantiation of a nested configstruct object."""

        @configstruct
        class MyConfig1:
            x: int
            y: str

        @configstruct
        class MyConfig2:
            v: MyConfig1
            w: Dict[str, MyConfig1]

        x_val1 = 1
        x_val2 = 2
        y_val1 = "hello"
        y_val2 = "world"
        my_config = MyConfig2(
            v=MyConfig1(x=x_val1, y=y_val1),
            w={"data": MyConfig1(x=x_val2, y=y_val2)}
        )

        self.assertEqual(my_config.v.x, x_val1)
        self.assertEqual(my_config.v.y, y_val1)
        self.assertEqual(my_config.w["data"].x, x_val2)
        self.assertEqual(my_config.w["data"].y, y_val2)

    def test_43_direct_instantiation_optional(self):
        """Test direct instantiation with optional fields."""

        @configstruct
        class MyConfig:
            x: int
            y: Optional[str] = None

        x_val = 3
        my_config = MyConfig(x=x_val)

        self.assertEqual(my_config.x, x_val)
        self.assertIsNone(my_config.y)

    def test_51_fail_direct_instantiation_wrong_type(self):
        """Fail on wrong field type."""

        @configstruct
        class MyConfig:
            x: int
            y: str

        with self.assertRaises(QMI_ConfigurationException):
            MyConfig(x=1, y=2)

    def test_52_fail_direct_instantiation_missing_field(self):
        """Fail on missing field."""

        @configstruct
        class MyConfig:
            x: int
            y: str

        with self.assertRaises(TypeError):
            MyConfig(x=1)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import Mock, MagicMock, patch
import os

import numpy

import qmi
from qmi.instruments.imagine_eyes import ImagineEyes_Mirao52e, MirrorStatus
from qmi.core.exceptions import QMI_InstrumentException

readline_mock = Mock()


class PopenMock:
    stdin = Mock(write=MagicMock())
    stdout = Mock(readline=readline_mock)

    def __init__(self, args="wine", stdin=None, stdout=None, cwd=None, env=None):
        pass

    def wait(self):
        pass


class InstanceCreationTestCase(unittest.TestCase):

    def test_default_directory(self):
        """Test that the behaviour is as expected when using default directory while creating instance."""
        if os.path.isdir(ImagineEyes_Mirao52e.DEFAULT_BRIDGE_DIR):
            expected_dir = ImagineEyes_Mirao52e.DEFAULT_BRIDGE_DIR
            mirao = ImagineEyes_Mirao52e(qmi.core.context.QMI_Context("Mirao"), "Mirao test")
            self.assertEqual(mirao._bridge_dir, expected_dir)

        else:
            with self.assertRaises(QMI_InstrumentException):
                ImagineEyes_Mirao52e(qmi.core.context.QMI_Context("Mirao"), "Mirao test")

    def test_custom_directory(self):
        """Test that a custom directory, set as local directory, does run without exceptions"""
        expected_dir = os.path.dirname(__file__)
        mirao = ImagineEyes_Mirao52e(qmi.core.context.QMI_Context("Mirao"), "Mirao test", expected_dir)
        self.assertEqual(mirao._bridge_dir, expected_dir)

    def test_faulty_custom_directory(self):
        """Test that a custom directory that does not exist raises an exception"""
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException):
            ImagineEyes_Mirao52e(qmi.core.context.QMI_Context("Mirao"), "Mirao test", r"Q:\utech\QID\QMI")


@patch("sys.platform", "linux1")
class MethodsLinuxTestCase(unittest.TestCase):

    @patch("sys.platform", "linux1")
    def setUp(self):
        # First mock is for the Linux WinE opener
        self._popen_patcher = patch("qmi.instruments.imagine_eyes.mirao52e.subprocess.Popen", new=PopenMock)
        self._popen_patcher.start()
        readline_mock.side_effect = [b"MIRAO52_BRIDGE bla", b"READY"] + [b"OK", b"READY"] * 3
        path = os.path.dirname(__file__)
        self.mirao = ImagineEyes_Mirao52e(qmi.core.context.QMI_Context("Mirao"), "Mirao test", path)

    @patch("sys.platform", "linux1")
    def tearDown(self):
        if self.mirao._is_open:
            self.mirao.close()

        self._popen_patcher.stop()

    def test_01_open(self):
        """Test opening the instrument in Linux."""
        self.mirao.open()
        self.assertTrue(self.mirao._is_open)

    def test_01b_open_excepts_at_handshake(self):
        """Test opening the instrument in Linux."""
        readline_mock.side_effect = [b"MIRAO52_BRIDGE", b"READY"] + [b"OK", b"READY"] * 3
        expected_exception = "Unexpected handshake from helper program: ['MIRAO52_BRIDGE']"
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.mirao.open()

        self.assertEqual(str(exc.exception), expected_exception)

    @patch("sys.platform", "linux1")
    def test_02_close(self):
        """Test closing the instrument in Linux."""
        self.mirao.open()
        self.mirao.close()
        self.assertFalse(self.mirao._is_open)

    def test_03_get_idn(self):
        """Test that the IDN request is correctly made and returns QMI_InstrumentIdentification instance."""
        expected_vendor = "ImagineEyes"
        expected_model = "Mirao52e"
        expected_version = "bla"

        self.mirao.open()
        val = self.mirao.get_idn()
        self.assertEqual(type(val), qmi.core.instrument.QMI_InstrumentIdentification)
        self.assertEqual(val.vendor, expected_vendor)
        self.assertEqual(val.model, expected_model)
        self.assertEqual(val.version, expected_version)
        self.assertEqual(self.mirao._version_string, expected_version)

    def test_04_apply(self):
        """See that the 'apply' command works as expected on Linux."""
        actuators = numpy.random.random(52)
        cmd = "APPLY " + ",".join("{:.6f}".format(v) for v in actuators) + "\r\n"
        expected_call = bytes(cmd.encode())
        self.mirao.open()
        self.mirao.apply(actuators)
        self.mirao._proc.stdin.write.assert_called_with(expected_call)

    def test_05_apply_wrong_actuator_length(self):
        """Test that the 'apply' excepts when length of the actuators is not 52."""
        actuators = numpy.random.random(51)
        self.mirao.open()
        with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
            self.mirao.apply(actuators)

        actuators = numpy.random.random(53)
        with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
            self.mirao.apply(actuators)

    def test_06_apply_erroneous_returns(self):
        """Test that the 'apply' excepts when program gives unvalid responses."""
        replies = [b"MIRAO52_BRIDGE bla", b"READY", b"OK", b"READY", b"ERROR Test Error", b"READY", b""]
        readline_mock.side_effect = replies + [b"READY", b"OK", b"READY"]
        expected_exception_1 = "Error from {}: {}".format("Mirao test", "Test Error")
        expected_exception_2 = "Unexpected end of input from helper program"
        expected_exception_3 = "Unexpected response from Mirao test: {}".format([])
        actuators = numpy.random.random(52)
        self.mirao.open()
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc1:
            self.mirao.apply(actuators)

        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc2:
            self.mirao.apply(actuators)

        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc3:
            self.mirao.apply(actuators)

        self.assertEqual(str(exc1.exception), expected_exception_1)
        self.assertEqual(str(exc2.exception), expected_exception_2)
        self.assertEqual(str(exc3.exception), expected_exception_3)

    def test_07_get_mirror_status(self):
        """See that 'get_mirror_status' returns correct MirrorStatus object."""
        psu_temp = 123.4
        locked = 0
        replies = [b"MIRAO52_BRIDGE bla", b"READY", b"OK", b"READY",
                   bytes(f"PSUTEMP={psu_temp} LOCKED={locked}".encode())]
        readline_mock.side_effect = replies + [b"OK", b"READY", b"OK", b"READY"]
        cmd = "MONITOR" + "\r\n"
        expected_call = bytes(cmd.encode())
        expected_result = MirrorStatus(psu_temp, bool(locked))
        self.mirao.open()
        mirror_status = self.mirao.get_mirror_status()
        self.mirao._proc.stdin.write.assert_called_with(expected_call)
        self.assertEqual(mirror_status, expected_result)


@patch("sys.platform", "win32")
class MethodsWindowsOnlyTestCase(unittest.TestCase):

    @patch("sys.platform", "win32")
    def setUp(self):
        self._popen_patcher = patch("qmi.instruments.imagine_eyes.mirao52e.subprocess.Popen", new=PopenMock)
        self._popen_patcher.start()
        readline_mock.side_effect = [b"MIRAO52_BRIDGE bla", b"READY"] + [b"OK", b"READY"] * 3
        path = os.path.dirname(__file__)
        self.mirao = ImagineEyes_Mirao52e(qmi.core.context.QMI_Context("Mirao"), "Mirao test", path)

    @patch("sys.platform", "win32")
    def tearDown(self):
        if self.mirao._is_open:
            self.mirao.close()

        self._popen_patcher.stop()

    def test_01_open(self):
        """Test opening the instrument in Windows."""
        self.mirao.open()
        self.assertTrue(self.mirao._is_open)

    def test_02_close(self):
        """Test closing the instrument in Windows."""
        self.mirao.open()
        self.mirao.close()
        self.assertFalse(self.mirao._is_open)

    def test_03_get_idn(self):
        """Test that the IDN request is correctly made and returns QMI_InstrumentIdentification instance."""
        expected_vendor = "ImagineEyes"
        expected_model = "Mirao52e"
        expected_version = "bla"

        self.mirao.open()
        val = self.mirao.get_idn()
        self.assertEqual(type(val), qmi.core.instrument.QMI_InstrumentIdentification)
        self.assertEqual(val.vendor, expected_vendor)
        self.assertEqual(val.model, expected_model)
        self.assertEqual(val.version, expected_version)
        self.assertEqual(self.mirao._version_string, expected_version)

    def test_04_apply(self):
        """See that the 'apply' command works as expected on Windows as well."""
        actuators = numpy.random.random(52)
        cmd = "APPLY " + ",".join("{:.6f}".format(v) for v in actuators) + "\r\n"
        expected_call = bytes(cmd.encode())
        self.mirao.open()
        self.mirao.apply(actuators)
        self.mirao._proc.stdin.write.assert_called_with(expected_call)

    def test_05_get_mirror_status(self):
        """See that 'get_mirror_status' returns correct MirrorStatus object."""
        psu_temp = 432.1
        locked = 1
        replies = [b"MIRAO52_BRIDGE bla", b"READY", b"OK", b"READY",
                   bytes(f"PSUTEMP={psu_temp} LOCKED={locked}".encode())]
        readline_mock.side_effect = replies + [b"OK", b"READY", b"OK", b"READY"]
        cmd = "MONITOR" + "\r\n"
        expected_call = bytes(cmd.encode())
        expected_result = MirrorStatus(psu_temp, bool(locked))
        self.mirao.open()
        mirror_status = self.mirao.get_mirror_status()
        self.mirao._proc.stdin.write.assert_called_with(expected_call)
        self.mirao._proc.stdout = Mock(readline=readline_mock)
        self.assertEqual(mirror_status, expected_result)


if __name__ == '__main__':
    unittest.main()



"""
Instrument driver for the Adwin.

This driver requires that the ADwin library and ADwin Python module
are installed on the system.

On Linux, the ADwin system library must be at least version 5.0.13.
Earlier versions do not correctly support double-precision floating point
data from the Adwin T12. On Windows, library version 6.0.x should be used.
"""

import logging
import os
import os.path
import sys
import threading
import time
import typing

from typing import Any, List, Union

import numpy as np

# Lazy import of the ADwin module. See the function _import_modules() below.
if typing.TYPE_CHECKING:
    import ADwin
else:
    ADwin = None

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# Global mutex to protect access to the Adwin library.
_adwin_mutex = threading.RLock()


def _import_modules() -> None:
    """Import the vendor-provided "ADwin" module.

    This import is done in a function, instead of at the top-level,
    to avoid an unnecessary dependency for programs that do not access
    the instrument directly.
    """
    global ADwin
    if ADwin is None:
        import ADwin  # pylint: disable=W0621


class Adwin_Base(QMI_Instrument):
    """Instrument driver base class for the Adwin real-time microcontrollers. ADwin.py has fixed maximum number of
    parameters and data that are dependent on processor type. The 'MAX_m' values here are valid for at least the T11,
    T12 and T12.1 processor types. The maximum number of slots instead is dependent on the enclosure type of the ADwin.
    """
    MAX_PAR: int = 80
    MAX_DATA: int = 200
    MAX_PROCESS_NO: int = 10

    # Maximum time to wait until an Adwin process stops (seconds).
    PROCESS_STOP_TIMEOUT = 1.0

    def __init__(self, context: QMI_Context, name: str, device_no: int) -> None:
        """Initialize the Adwin driver.

        Parameters:
            context: QMI context.
            name: Name for this instrument instance.
            device_no: ADwin device number.
        """
        super().__init__(context, name)

        _logger.info("[%s] Creating Adwin instance, device_no=%d", name, device_no)
        self._device_no = device_no

        _import_modules()

        # NOTE: Creating the ADwin instance does not yet initiate communication with the device.
        self._adwin = ADwin.ADwin(DeviceNo=self._device_no, raiseExceptions=1)

    def _is_integer_array(self, data_idx: int) -> bool:
        """Determine whether the specified global data array has integer elements.

        This function must be called with `_adwin_mutex` locked.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).

        Returns:
            True if the array has integer elements, False if it has floating point elements.

        Raise:
            QMI_InstrumentException: If the array has neither integer, nor floating point elements.
        """
        # See "Python_eng.pdf" page 30 for how "type_code" and "type_name" encode data types.
        type_code, type_name = self._adwin.Data_Type(data_idx)

        # Check on "type_name", since not all "type_code" values are enumerated in the ADwin class.
        if type_name in ("byte", "short", "int32", "int64", "long"):
            return True
        elif type_name in ("float", "float32", "float64"):  # "float" for backwards compatibility (T11 and older)
            return False
        else:
            # "string" or "undefined".
            raise QMI_InstrumentException("Unsupported data type {} ({}) for Data {}".format(
                type_name, type_code, data_idx
            ))

    @rpc_method
    def reboot(self) -> None:
        """Reboot the Adwin system, clear data and restart programs.

        This function must be called after power-on before communication
        with the Adwin is possible. This function may be called again
        to perform a full reset.

        This function boots the Adwin, removes all processes and data.
        """
        self._check_is_open()

        # Find boot loader file.
        with _adwin_mutex:
            processor_type = self._adwin.Processor_Type().strip()
            boot_file = "ADwin{}.btl".format(processor_type.replace(".", "").strip("T"))
            boot_file_path = os.path.join(self._adwin.ADwindir, boot_file)

        if not os.path.isfile(boot_file_path):
            # On Linux, boot files are in a subdirectory under the ADwindir.
            boot_file_path = os.path.join(self._adwin.ADwindir, "share", "btl", boot_file)

        if not os.path.isfile(boot_file_path):
            raise QMI_InstrumentException("Can not find boot file {}".format(boot_file))

        _logger.info("[%s] Rebooting Adwin", self._name)
        with _adwin_mutex:
            self._adwin.Boot(boot_file_path)

        self.check_ready()

    @rpc_method
    def check_ready(self) -> None:
        """Raise an exception if the Adwin is not ready."""
        self._check_is_open()
        with _adwin_mutex:
            err = self._adwin.Test_Version()

        if err != 0:
            raise QMI_InstrumentException("Incorrect Adwin operating system version {}".format(err))

    @rpc_method
    def get_processor_type(self) -> str:
        """Returns the ADwin processor type."""
        self._check_is_open()
        with _adwin_mutex:
            return self._adwin.Processor_Type()

    @rpc_method
    def get_workload(self) -> int:
        """Return average processor workload in percent since the last call."""
        self._check_is_open()
        with _adwin_mutex:
            return self._adwin.Workload()

    @rpc_method
    def load_process(self, bin_file: str) -> None:
        """Load the specified Adwin process but do not start it yet.

        Parameters:
            bin_file: Path to binary file to load into the Adwin.
        """
        self._check_is_open()
        with _adwin_mutex:
            self._adwin.Load_Process(bin_file)

    @rpc_method
    def start_process(self, slot: int) -> None:
        """Start the specified Adwin process.

        The process must already have been loaded by calling `load_process()`.

        Parameters:
            slot: Adwin process slot index (range 1 to MAX_PROCESS_NO).
        """
        if (slot < 1) or (slot > self.MAX_PROCESS_NO):
            raise ValueError("Invalid process slot index")
        self._check_is_open()
        with _adwin_mutex:
            status = self._adwin.Process_Status(slot)
            if status != 0:
                raise QMI_InstrumentException("Can not start Adwin process, slot {} already running".format(slot))
            self._adwin.Start_Process(slot)

    @rpc_method
    def stop_process(self, slot: int) -> None:
        """Stop the specified Adwin process.

        Parameters:
            slot: Adwin process slot index (range 1 to MAX_PROCESS_NO).
        """
        if (slot < 1) or (slot > self.MAX_PROCESS_NO):
            raise ValueError("Invalid process slot index")
        self._check_is_open()
        with _adwin_mutex:
            self._adwin.Stop_Process(slot)

    @rpc_method
    def wait_for_process(self, slot: int, timeout: float) -> None:
        """Wait until the specified Adwin process has stopped.

        Parameters:
            slot: Adwin process slot index (range 1 to MAX_PROCESS_NO).
            timeout: Maximum time to wait (seconds).

        Raises:
            ~qmi.core.exceptions.QMI_TimeoutException: If the process does not stop before timeout occurs.
        """
        if (slot < 1) or (slot > self.MAX_PROCESS_NO):
            raise ValueError("Invalid process slot index")
        self._check_is_open()
        t0 = time.monotonic()
        while True:
            with _adwin_mutex:
                status = self._adwin.Process_Status(slot)
            if status == 0:
                break
            t = time.monotonic()
            if t - t0 > timeout:
                raise QMI_TimeoutException("Adwin process slot {} still running".format(slot))
            time.sleep(0.02)

    @rpc_method
    def is_process_running(self, slot: int) -> bool:
        """Return True if the specified Adwin process is currently running.

        Parameters:
            slot: Adwin process slot index (range 1 to MAX_PROCESS_NO).
        """
        if (slot < 1) or (slot > self.MAX_PROCESS_NO):
            raise ValueError("Invalid process slot index")
        self._check_is_open()
        with _adwin_mutex:
            status = self._adwin.Process_Status(slot)
        return (status != 0)

    @rpc_method
    def get_par(self, par_idx: int) -> int:
        """Return the current value of the specified global parameter.

        Parameters:
            par_idx: Index into the global Par array (range 1 to MAX_PAR).

        Returns:
            Value of the parameter.
        """
        if (par_idx < 1) or (par_idx > self.MAX_PAR):
            raise ValueError("Invalid Par index {}".format(par_idx))
        self._check_is_open()
        with _adwin_mutex:
            return self._adwin.Get_Par(par_idx)

    @rpc_method
    def get_fpar(self, par_idx: int) -> float:
        """Return the current value of the specified global parameter.

        Parameters:
            par_idx: Index into the global FPar array (range 1 to MAX_PAR).

        Returns:
            Value of the parameter.
        """
        if (par_idx < 1) or (par_idx > self.MAX_PAR):
            raise ValueError("Invalid FPar index {}".format(par_idx))
        self._check_is_open()
        with _adwin_mutex:
            return self._adwin.Get_FPar_Double(par_idx)

    @rpc_method
    def get_par_block(self, par_first: int, par_count: int) -> np.ndarray:
        """Return the current values of a block of global parameters.

        Parameters:
            par_first: First index into the global Par array (range 1 to MAX_PAR).
            par_count: Number of parameters to get.

        Returns:
            1D Numpy array with the parameter values.
        """
        if (par_first < 1) or (par_count < 1) or (par_first + par_count - 1 > self.MAX_PAR):
            raise ValueError("Invalid Par index range {} .. {}".format(par_first, par_first + par_count - 1))
        self._check_is_open()
        with _adwin_mutex:
            value = self._adwin.Get_Par_Block(par_first, par_count)
        return np.array(value)

    @rpc_method
    def get_fpar_block(self, par_first: int, par_count: int) -> np.ndarray:
        """Return the current values of a block of global parameters.

        Parameters:
            par_first: First index into the global FPar array (range 1 to MAX_PAR).
            par_count: Number of parameters to get.

        Returns:
            1D Numpy array with the parameter values.
        """
        if (par_first < 1) or (par_count < 1) or (par_first + par_count - 1 > self.MAX_PAR):
            raise ValueError("Invalid FPar index range {} .. {}".format(par_first, par_first + par_count - 1))
        self._check_is_open()
        with _adwin_mutex:
            value = self._adwin.Get_FPar_Block_Double(par_first, par_count)
        return np.array(value)

    @rpc_method
    def set_par(self, par_idx: int, value: int) -> None:
        """Change the value of the specified global parameter.

        Parameters:
            par_idx: Index into the global Par array (range 1 to MAX_PAR).
            value: New value to write to the parameter (signed 32-bit integer).
        """
        if (par_idx < 1) or (par_idx > self.MAX_PAR):
            raise ValueError("Invalid Par index {}".format(par_idx))
        if not isinstance(value, int):
            raise TypeError("Expecting integer value for parameter but got {}".format(value))
        self._check_is_open()
        with _adwin_mutex:
            self._adwin.Set_Par(par_idx, value)

    @rpc_method
    def set_fpar(self, par_idx: int, value: float) -> None:
        """Change the value of the specified global parameter.

        Parameters:
            par_idx: Index into the global FPar array (range 1 to MAX_PAR).
            value: New value to write to the parameter.
        """
        if (par_idx < 1) or (par_idx > self.MAX_PAR):
            raise ValueError("Invalid FPar index {}".format(par_idx))
        if not isinstance(value, (int, float)):
            raise TypeError("Expecting numeric value for parameter but got {}".format(value))
        self._check_is_open()
        with _adwin_mutex:
            self._adwin.Set_FPar_Double(par_idx, value)

    @rpc_method
    def get_data_length(self, data_idx: int) -> int:
        """Return the length of the specified global Data array.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).

        Returns:
            Length of the data array (maximum valid array index).
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        self._check_is_open()
        with _adwin_mutex:
            return self._adwin.Data_Length(data_idx)

    @rpc_method
    def get_data(self, data_idx: int, first_index: int, count: int) -> np.ndarray:
        """Read values from a global Data array.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).
            first_index: First index into the data array (>= 1).
            count: Number of elements to read from the data array.

        Returns:
            1D Numpy array with data values.
            Element 0 of the returned array will correspond with `first_index` into the Adwin array.
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        if (first_index < 1) or (count < 1):
            raise ValueError("Invalid Data element index range {} .. {}".format(first_index, first_index + count - 1))
        self._check_is_open()
        with _adwin_mutex:
            if self._is_integer_array(data_idx):
                value = self._adwin.GetData_Long(data_idx, first_index, count)
            else:
                value = self._adwin.GetData_Double(data_idx, first_index, count)
        return np.array(value)

    @rpc_method
    def get_full_data(self, data_idx: int) -> np.ndarray:
        """Read all elements from a global DATA array.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).

        Returns:
            1D Numpy array with data values.
            Element 0 of the returned array will correspond with element 1 into the Adwin array.
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        self._check_is_open()
        with _adwin_mutex:
            length = self._adwin.Data_Length(data_idx)
            if self._is_integer_array(data_idx):
                value = self._adwin.GetData_Long(data_idx, 1, length)
            else:
                value = self._adwin.GetData_Double(data_idx, 1, length)
        return np.array(value)

    @rpc_method
    def set_data(self, data_idx: int, first_index: int, value: Union[np.ndarray, List[int], List[float]]) -> None:
        """Write values to a global Data array.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).
            first_index: First index into the Data array (>= 1).
            value: List of values to write to the Data array.
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        if first_index < 1:
            raise ValueError("Invalid Data element index {}".format(first_index))
        self._check_is_open()
        with _adwin_mutex:
            raw_value: Any
            if self._is_integer_array(data_idx):
                if isinstance(value, np.ndarray):
                    # The Adwin library does not accept Numpy arrays.
                    if value.dtype.kind not in "iu":
                        raise ValueError("Invalid non-integer data for Data_{}".format(data_idx))
                    raw_value = np.ctypeslib.as_ctypes(value.astype(np.int32))
                else:
                    raw_value = value
                self._adwin.SetData_Long(raw_value, data_idx, first_index, len(value))
            else:
                if isinstance(value, np.ndarray):
                    # The Adwin library does not accept Numpy arrays.
                    raw_value = np.ctypeslib.as_ctypes(value.astype(np.float64))
                else:
                    raw_value = value
                self._adwin.SetData_Double(raw_value, data_idx, first_index, len(value))

    @rpc_method
    def get_fifo_filled(self, data_idx: int) -> int:
        """Return the number of values waiting in the FIFO.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).

        Returns:
            Number of FIFO entries currently in use.
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        self._check_is_open()
        with _adwin_mutex:
            return self._adwin.Fifo_Full(data_idx)

    @rpc_method
    def get_fifo_room(self, data_idx: int) -> int:
        """Return the number of values that can be added to the FIFO without overflowing it.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).

        Returns:
            Number of empty FIFO entries available.
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        self._check_is_open()
        with _adwin_mutex:
            return self._adwin.Fifo_Empty(data_idx)

    @rpc_method
    def read_fifo(self, data_idx: int, count: int) -> np.ndarray:
        """Read data from a global FIFO variable.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).
            count: Number of values to read from the FIFO.

        Returns:
            1D Numpy array containing the FIFO elements.
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        self._check_is_open()
        with _adwin_mutex:
            if self._is_integer_array(data_idx):
                value = self._adwin.GetFifo_Long(data_idx, count)
            else:
                value = self._adwin.GetFifo_Double(data_idx, count)
        return np.array(value)

    @rpc_method
    def write_fifo(self, data_idx: int, value: Union[np.ndarray, List[int], List[float]]) -> None:
        """Write data to a global FIFO variable.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).
            value: Array of values to write to the FIFO.
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        self._check_is_open()
        with _adwin_mutex:
            raw_value: Any
            if self._is_integer_array(data_idx):
                if isinstance(value, np.ndarray):
                    # The Adwin library does not accept Numpy arrays.
                    raw_value = np.ctypeslib.as_ctypes(value.astype(np.int32))
                    if value.dtype.kind not in "iu":
                        raise ValueError("Invalid non-integer data for Data_{}".format(data_idx))
                else:
                    raw_value = value
                self._adwin.SetFifo_Long(data_idx, raw_value, len(value))
            else:
                if isinstance(value, np.ndarray):
                    # The Adwin library does not accept Numpy arrays.
                    raw_value = np.ctypeslib.as_ctypes(value.astype(np.float64))
                else:
                    raw_value = value
                self._adwin.SetFifo_Double(data_idx, raw_value, len(value))

    @rpc_method
    def set_file_to_data(self, data_idx: int, first_index: int, file_path: str) -> None:
        """Upload the file at `file_path` to the specified array.
        This method is used to load binary programs into TiCo modules at ADwin runtime.

        Parameters:
            data_idx: Index into the global list of Data arrays (range 1 to MAX_DATA).
            first_index: First index into the Data array (>= 1).
            file_path: Path of the file to upload.
        """
        if (data_idx < 1) or (data_idx > self.MAX_DATA):
            raise ValueError("Invalid Data index {}".format(data_idx))
        if first_index < 1:
            raise ValueError("Invalid Data element index {}".format(first_index))
        self._check_is_open()
        with _adwin_mutex:
            if not self._is_integer_array(data_idx):
                _logger.error("Binary files should be uploaded to long arrays.")
                raise QMI_InstrumentException("Target data array should be of type long.")

            # Here we workaround a bug in the ADwin library, data type should be
            # two, but on Linux it needs 3 to work correctly. They are working
            # on it. Matteo Pompili June 2021.
            if sys.platform.startswith("linux"):
                adwin_data_type = 3  # Long array
            else:
                adwin_data_type = 2  # Long array

            # Check if file exists, so we can provide a better error than what ADwin would.
            if not os.path.isfile(file_path):
                _logger.error("Trying to upload %s to the ADwin, but it does not exist.", file_path)
                raise QMI_InstrumentException("The specified file does not exist.")

            self._adwin.File2Data(file_path, adwin_data_type, data_idx, first_index)



"""
PicoQuant, time tagger instruments and event handling.

The qmi.instruments.pi package provides support for:
- MultiHarp 150
- HydraHarp 400
- PicoHarp 300.
"""
from qmi.instruments.picoquant.support._decoders import EventDataType, EventFilterMode, SYNC_TYPE
from qmi.instruments.picoquant.support._realtime import RealTimeHistogram, RealTimeCountRate, TttrHistogram
from qmi.instruments.picoquant.hydraharp import PicoQuant_HydraHarp400
from qmi.instruments.picoquant.multiharp import PicoQuant_MultiHarp150
from qmi.instruments.picoquant.picoharp import PicoQuant_PicoHarp300



"""Test for the NKT Photonics boostik driver."""

import logging

from unittest import TestCase
from unittest.mock import MagicMock
from unittest.mock import patch
from unittest.mock import call

from typing import cast

from dataclasses import dataclass


from qmi.core.transport import QMI_Transport
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.instruments.nkt_photonics.boostik import KoherasBoostikLaserAmplifier
from qmi.core.exceptions import QMI_InstrumentException


# Disable all logging
logging.disable(logging.CRITICAL)


@dataclass
class TestMeta:
    """Test meta data."""

    transport: MagicMock
    transport_str: MagicMock
    name: MagicMock
    super: MagicMock
    instr: KoherasBoostikLaserAmplifier


class TestKoherasBoostikLaserAmplifier(TestCase):
    """Testcase for the KoherasBoostikLaserAmplifier class."""

    def setUp(self):
        mock_transport = MagicMock(spec=QMI_Transport)
        mock_name = MagicMock()
        mock_transport_str = MagicMock()
        mock_super = MagicMock()

        with patch(
            "qmi.instruments.nkt_photonics.boostik.create_transport",
            return_value=mock_transport,
        ):
            instr = KoherasBoostikLaserAmplifier(
                MagicMock(), mock_name, mock_transport_str
            )

        self._patcher_super = patch(
            "qmi.instruments.nkt_photonics.boostik.super", mock_super
        )
        self._patcher_super.start()

        self._meta = TestMeta(
            transport=mock_transport,
            instr=cast(KoherasBoostikLaserAmplifier, instr),
            transport_str=mock_transport_str,
            name=mock_name,
            super=mock_super,
        )

    def tearDown(self):
        self._meta = None
        self._patcher_super.stop()

    def test_init(self):
        """KoherasBoostikLaserAmplifier.__init__(), happy flow."""
        self.assertEqual(self._meta.instr._transport, self._meta.transport)

    def test_open(self):
        """KoherasBoostikLaserAmplifier.open(), happy flow."""

        self._meta.instr.open()
        self._meta.transport.open.assert_called_once_with()
        self._meta.transport.discard_read.assert_called_once_with()
        self._meta.super().open.assert_called_once_with()

    def test_close(self):
        """KoherasBoostikLaserAmplifier.close(), happy flow."""

        self._meta.instr.close()
        self._meta.transport.close.assert_called_once_with()
        self._meta.super().close.assert_called_once_with()

    def test_send_command(self):
        """KoherasBoostikLaserAmplifier._send_command(), happy flow."""

        mock_cmd = MagicMock()
        self._meta.instr._send_command(mock_cmd)
        self._meta.transport.write.assert_called_once_with(
            (mock_cmd + "\r\n").encode("ascii")
        )

    def test_get_string(self):
        """KoherasBoostikLaserAmplifier._get_string(), happy flow."""

        mock_cmd = MagicMock()
        mock_resp = "RESPONSE\r\n".encode("ascii")

        self._meta.instr._send_command = MagicMock()
        self._meta.transport.read_until = MagicMock(return_value=mock_resp)
        rt_val = self._meta.instr._get_string(mock_cmd)

        self._meta.instr._send_command.assert_called_once_with(mock_cmd)
        self.assertEqual(rt_val, "RESPONSE")

    def _test_get_string_conversion(self, function_name, resp):
        mock_cmd = MagicMock()
        mock_resp = resp

        self._meta.instr._get_string = MagicMock(return_value=mock_resp)
        rt_val = getattr(self._meta.instr, function_name)(mock_cmd)
        self.assertEqual(rt_val, mock_resp)

    def test_get_float(self):
        """KoherasBoostikLaserAmplifier._get_float(), happy flow."""
        self._test_get_string_conversion("_get_float", 0.005)

    def test_get_integer(self):
        """KoherasBoostikLaserAmplifier._get_integer(), happy flow."""
        self._test_get_string_conversion("_get_integer", 5)

    def test_get_bool(self):
        """KoherasBoostikLaserAmplifier._get_boolean(), happy flow."""
        self._test_get_string_conversion("_get_boolean", True)

    def _test_get_function(self, function_name, get_function_name, cmd):
        mock_resp = MagicMock()
        setattr(self._meta.instr, get_function_name, MagicMock(return_value=mock_resp))
        rt_val = getattr(self._meta.instr, function_name)()
        self.assertEqual(rt_val, mock_resp)
        getattr(self._meta.instr, get_function_name).assert_called_once_with(cmd)

    def test_get_current_setpoint(self):
        """KoherasBoostikLaserAmplifier.get_current_setpoint(), happy flow."""
        self._test_get_function("get_current_setpoint", "_get_float", "ACC")

    def test_get_actual_current(self):
        """KoherasBoostikLaserAmplifier.get_actual_current(), happy flow."""
        self._test_get_function("get_actual_current", "_get_float", "AMC")

    def test_get_diode_booster_temperature(self):
        """KoherasBoostikLaserAmplifier.get_diode_booster_temperature(), happy flow."""
        self._test_get_function("get_diode_booster_temperature", "_get_float", "AMT 1")

    def test_get_ambient_temperature(self):
        """KoherasBoostikLaserAmplifier.get_ambient_temperature(), happy flow."""
        self._test_get_function("get_ambient_temperature", "_get_float", "CMA")

    def test_get_input_power(self):
        """KoherasBoostikLaserAmplifier.get_input_power(), happy flow."""
        self._test_get_function("get_input_power", "_get_float", "CMP 1")

    def test_get_amplifier_enabled(self):
        """KoherasBoostikLaserAmplifier.get_amplifier_enabled(), happy flow."""
        self._test_get_function("get_amplifier_enabled", "_get_boolean", "CDO")

    def test_get_amplifier_information(self):
        """KoherasBoostikLaserAmplifier.get_amplifier_information(), happy flow."""
        self._test_get_function("get_amplifier_information", "_get_string", "CDI")

    def test_set_current_setpoint(self):
        """KoherasBoostikLaserAmplifier.set_current_setpoint(), happy flow."""
        mock_value = MagicMock()
        mock_resp = MagicMock()
        self._meta.instr._get_float = MagicMock(return_value=mock_resp)

        rt_val = self._meta.instr.set_current_setpoint(mock_value)
        self.assertEqual(rt_val, mock_resp)
        self._meta.instr._get_float.assert_called_once_with(f"ACC {mock_value}")

    def test_set_amplifier_enabled(self):
        """KoherasBoostikLaserAmplifier.set_amplifier_enabled(), happy flow."""
        mock_value = MagicMock()
        mock_resp = MagicMock()
        self._meta.instr._get_boolean = MagicMock(return_value=mock_resp)

        rt_val = self._meta.instr.set_amplifier_enabled(mock_value)
        self.assertEqual(rt_val, mock_resp)
        self._meta.instr._get_boolean.assert_called_once_with(
            f"CDO {int(bool(mock_value))}"
        )



#! /usr/bin/env python3

"""Test QMI task functionality."""

import logging
import random
import threading
import time
import unittest
from collections import namedtuple
from unittest.mock import sentinel

import qmi
import qmi.core.exceptions
from qmi.utils.context_managers import start_stop, start_stop_join
from qmi.core.pubsub import QMI_Signal, QMI_SignalReceiver
from qmi.core.rpc import QMI_RpcObject, rpc_method
from qmi.core.task import (
    QMI_Task,
    QMI_LoopTask,
    QMI_LoopTaskMissedLoopPolicy,
    QMI_TaskRunner,
)

from tests.patcher import PatcherQmiContext


class CustomTaskRunner(QMI_TaskRunner):
    @rpc_method
    def get_custom_taskrunner_attr(self):
        return sentinel.custom_taskrunner_attr


class SimpleTestTask(QMI_Task):
    """A simple test task.

    It waits for a configurable time, then emits a signal and stops.
    Alternatively, the task can be configured to raise an exception during initialization or while running.
    """

    Settings = namedtuple("Settings", "message value")

    my_signal = QMI_Signal([int])

    def __init__(
        self,
        task_runner,
        name,
        raise_exception_in_init,
        raise_exception_in_run,
        duration,
        value,
    ):
        super().__init__(task_runner, name)

        self.settings = SimpleTestTask.Settings(
            "Hello, world from within SimpleTestTask!", value
        )
        self.raise_exception_in_run = raise_exception_in_run
        self.duration = duration

        if raise_exception_in_init:
            raise ValueError(
                "This error originates from within SimpleTestTask's __init__() method."
            )

    def run(self):
        """Run this task."""

        if self.raise_exception_in_run:
            raise ValueError(
                "This error originates from within SimpleTestTask's run() method."
            )

        run_until = time.monotonic() + self.duration
        while time.monotonic() < run_until:
            self.update_settings()
            self.sleep(0.100)  # Wait for 100 ms.

        self.status = self.settings.value
        self.my_signal.publish(self.settings.value)


class SlowTestTask(SimpleTestTask):
    """A slow version of simple test task.

    It waits for a configurable time, then emits a signal and stops.
    The difference to the SimpleTestTask is that the settings update rate is much slower.
    """

    Settings = namedtuple("Settings", "message value")

    my_signal = QMI_Signal([int])

    def __init__(self, task_runner, name, duration, value, event):
        super().__init__(task_runner, name, False, False, duration, value)
        self.event = event

    def run(self):
        """Run this task."""

        run_until = time.monotonic() + self.duration
        while time.monotonic() < run_until:
            if self.event.is_set():
                self.update_settings()
                self.event.clear()

            self.sleep(0.1)

        self.status = self.settings.value
        self.my_signal.publish(self.settings.value)


class TestPublisher(QMI_RpcObject):
    """A simple RPC object which can publish a signal."""

    sig_test = QMI_Signal([int])

    @rpc_method
    def send_signal(self, value):
        self.sig_test.publish(value)


class SignalWaitingTask(QMI_Task):
    """A test task that waits for signals.

    This task subscribes to signals from another object.
    The task waits for a signal, then stops or continues depending on
    the contents of the signal.
    """

    def __init__(self, task_runner, name, wait_timeout, context_id):
        super().__init__(task_runner, name)
        self.wait_timeout = wait_timeout
        self.receiver = QMI_SignalReceiver()
        publisher_proxy = qmi.get_task(f"{context_id}.test_publisher")
        publisher_proxy.sig_test.subscribe(self.receiver)
        self.status = None

    def run(self):
        while True:
            try:
                signal = self.receiver.get_next_signal(timeout=self.wait_timeout)
            except qmi.core.exceptions.QMI_TaskStopException:
                # This happens if the task is stopped while waiting for a signal.
                self.status = -100
                raise
            except qmi.core.exceptions.QMI_TimeoutException:
                # This happens if no signal arrives within `wait_timeout`.
                self.status = -101
                raise
            assert signal.signal_name == "sig_test"
            (value,) = signal.args
            if value == -1:
                break
            self.status = value


class LoopTestTask(QMI_LoopTask):
    """A loop test task.

    This task overrides the standard QMI_LoopTask methods and tests them. It also tests that the signals get
    published when changed, and that a loop time-outing its iteration period are handled as expected from the attribute
    set for `QMI_LoopTaskMissedLoopPolicy`.
    """

    Status = namedtuple("Status", "message value")
    Settings = namedtuple("Settings", "message value")

    my_signal = QMI_Signal([int])

    def __init__(
        self,
        task_runner,
        name,
        status_value,
        loop_period,
        policy,
        nr_of_loops,
        increase_loop,
        error=None,
    ):
        super().__init__(task_runner, name, loop_period, policy)
        # Initialize settings and status with some (default) values
        self.settings = LoopTestTask.Settings("Standard setting for LoopTestTask", 1)
        self.status = LoopTestTask.Status("LoopTestTask loop iteration status is", status_value)
        self._status_value = status_value
        self._nr_of_loops = nr_of_loops
        self._increase_loop = increase_loop
        self._error = error
        self._current_loop = 0

    def loop_prepare(self):
        # Initial settings and status value publishing
        self.sig_settings_updated.publish(self.settings.value)
        self.sig_status_updated.publish(self.status.value)
        # sleep until next monotonic period start. Makes sure we won't get random missed loops depending on which
        # monotonic moment we start our looping
        time_to_sleep = self._loop_period - (time.monotonic() % self._loop_period)
        if time_to_sleep > 0:
            self.sleep(time_to_sleep)

    def loop_iteration(self):
        # This follows settings update and processing step if settings are changed. Status is updated after this.
        self._current_loop += 1  # increase loop value
        # do some kind of "work"
        self._status_value += 1
        self.status = LoopTestTask.Status("LoopTestTask loop iteration status is", self._status_value)
        self.settings = LoopTestTask.Settings("Standard setting for LoopTestTask", self._current_loop)
        if self._current_loop >= self._nr_of_loops:
            time.sleep(self._loop_period)  # Make sure the latest loop finishes before stopping
            self._task_runner.stop()
            return

        if self._increase_loop:
            self._status_value += 1  # increase status value
            time.sleep(self._loop_period / 2.0)

        if self._error:
            time.sleep(self._loop_period * 1.1)  # Cause time-out in looping
            if self._error == "SKIP":
                self._current_loop += 1  # skip a loop

    def loop_finalize(self):
        # Clean up at the end of the loop task and set status and settings back to default (1)
        time.sleep(self._loop_period / 2.0)
        self.status = LoopTestTask.Status("LoopTestTask loop iteration status is", 1)
        self.settings = LoopTestTask.Settings("Standard setting for LoopTestTask", 1)
        self._task_runner.set_settings(self.settings)

    def update_status(self):
        # Try to obtain new status.
        if self._status_fifo:
            self.status = self._status_fifo.pop()
            return True

        return False

    def process_new_settings(self) -> None:
        # Update the settings
        self.settings = LoopTestTask.Settings(
            "Standard setting for LoopTestTask", self._current_loop
        )

    def publish_signals(self):
        # Publish something
        self.my_signal.publish(self.settings.value + self.status.value)
        self.sig_settings_updated.publish(self.settings.value)
        self.sig_status_updated.publish(self.status.value)


class TestQMITaskContextManager(unittest.TestCase):
    """Test the various context managers."""
    def test_with_context_manager(self):
        """Test the 'with' context manager run as QMI_TaskRunner."""
        logging.getLogger("qmi.core.task").setLevel(logging.ERROR)
        logging.getLogger("qmi.core.rpc").setLevel(logging.ERROR)
        qmi.start("test-taskrunner")
        with qmi.make_task(
            "taskrunner", SimpleTestTask, False, False, 1.0, 2.0
        ) as task:
            task.get_status()

        qmi.stop()
        logging.getLogger("qmi.core.task").setLevel(logging.NOTSET)
        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)

    def test_start_stop_context_manager(self):
        """Test the 'start_stop' context manager run as QMI_TaskRunner."""
        logging.getLogger("qmi.core.task").setLevel(logging.ERROR)
        logging.getLogger("qmi.core.rpc").setLevel(logging.ERROR)
        qmi.start("test-taskrunner")
        task: QMI_TaskRunner = qmi.make_task(
            "taskrunner", SimpleTestTask, False, False, 1.0, 2.0
        )
        with start_stop(task):
            task.get_status()

        task.join()
        qmi.stop()
        logging.getLogger("qmi.core.task").setLevel(logging.NOTSET)
        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)

    def test_start_stop_join_context_manager(self):
        """Test the 'start_stop_join' context manager run as QMI_TaskRunner."""
        logging.getLogger("qmi.core.task").setLevel(logging.ERROR)
        logging.getLogger("qmi.core.rpc").setLevel(logging.ERROR)
        qmi.start("test-taskrunner2")
        task: QMI_TaskRunner = qmi.make_task(
            "taskrunner2", SimpleTestTask, False, False, 1.0, 2.0
        )
        with start_stop_join(task):
            task.get_status()

        qmi.stop()
        logging.getLogger("qmi.core.task").setLevel(logging.NOTSET)
        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)


class TestQMITasks(unittest.TestCase):
    def setUp(self):
        logging.getLogger("qmi.core.task").setLevel(logging.ERROR)
        logging.getLogger("qmi.core.rpc").setLevel(logging.ERROR)
        self._ctx_qmi_id = f"test-tasks-{random.randint(0, 100)}"
        self.qmi_patcher = PatcherQmiContext()
        self.qmi_patcher.start(self._ctx_qmi_id)

    def tearDown(self):
        self.qmi_patcher.stop()
        logging.getLogger("qmi.core.task").setLevel(logging.NOTSET)
        logging.getLogger("qmi.core.rpc").setLevel(logging.NOTSET)

    def test_context_manager(self):
        """Test the 'with' context manager does the same as start_stop_join context manager."""
        with qmi.make_task(
                "simple_task_init",
                SimpleTestTask,
                raise_exception_in_init=False,
                raise_exception_in_run=False,
                duration=1.0,
                value=5,
        ) as simple_task:
            self.assertTrue(simple_task.is_running())

        self.assertFalse(simple_task.is_running())

    def test_exception_double_start(self):
        """Test QMI_UsageException is raised if the task is started twice."""
        with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
            task_proxy = qmi.make_task(
                "simple_task_init",
                SimpleTestTask,
                raise_exception_in_init=False,
                raise_exception_in_run=False,
                duration=1.0,
                value=5,
            )
            task_proxy.start()
            task_proxy.start()

        # Remove task proxy and re-run within context manager.
        qmi.context().remove_rpc_object(task_proxy)

        with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
            with qmi.make_task(
                "simple_task_init",
                SimpleTestTask,
                raise_exception_in_init=False,
                raise_exception_in_run=False,
                duration=1.0,
                value=5,
            ) as task_proxy:
                task_proxy.start()

    def test_exception_duplicate_task(self):
        """Test QMI_UsageException is raised if the task is created twice."""
        task_proxy = qmi.make_task(
            "simple_task_init",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=1.0,
            value=5,
        )

        with self.assertRaises(qmi.core.exceptions.QMI_DuplicateNameException):
            with qmi.make_task(
                "simple_task_init",
                SimpleTestTask,
                raise_exception_in_init=False,
                raise_exception_in_run=False,
                duration=1.0,
                value=5,
            ) as task_proxy:
                pass

    def test_exception_during_init(self):
        """Test QMI_TaskInitException is raised."""
        with self.assertRaises(qmi.core.exceptions.QMI_TaskInitException):
            qmi.make_task(
                "simple_task_init",
                SimpleTestTask,
                raise_exception_in_init=True,
                raise_exception_in_run=False,
                duration=1.0,
                value=5,
            )

    def test_exception_during_run(self):
        """Test QMI_TaskRunException is raised."""
        task_proxy = qmi.make_task(
            "simple_task_run",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=True,
            duration=1.0,
            value=6,
        )
        task_proxy.start()
        time.sleep(2.0)
        with self.assertRaises(qmi.core.exceptions.QMI_TaskRunException):
            task_proxy.join()

    def test_exception_during_context_run(self):
        """Test QMI_TaskRunException is raised within context."""
        with self.assertRaises(qmi.core.exceptions.QMI_TaskRunException):
            with qmi.make_task(
                "simple_task_run",
                SimpleTestTask,
                raise_exception_in_init=False,
                raise_exception_in_run=True,
                duration=1.0,
                value=6,
            ) as task_proxy:
                time.sleep(2.0)

        self.assertFalse(task_proxy.is_running())

    def test_run_to_completion(self):
        """Test task stops running after completion of the task."""
        task_proxy = qmi.make_task(
            "simple_task_complete",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=1.0,
            value=7,
        )
        self.assertFalse(task_proxy.is_running())
        task_proxy.start()
        self.assertTrue(task_proxy.is_running())
        time.sleep(2.0)
        self.assertFalse(task_proxy.is_running())
        t1 = time.monotonic()
        task_proxy.join()
        t2 = time.monotonic()
        self.assertLess(t2 - t1, 0.1)
        status = task_proxy.get_status()
        self.assertEqual(status, 7)

    def test_run_to_stopped(self):
        """Test task stops running if it is called to stop."""
        task_proxy = qmi.make_task(
            "simple_task_stopped",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=2.0,
            value=8,
        )
        task_proxy.start()
        time.sleep(0.5)
        self.assertTrue(task_proxy.is_running())
        t1 = time.monotonic()
        task_proxy.stop()
        t2 = time.monotonic()
        self.assertLess(t2 - t1, 0.1)
        time.sleep(0.1)
        self.assertFalse(task_proxy.is_running())
        t1 = time.monotonic()
        task_proxy.join()
        t2 = time.monotonic()
        self.assertLess(t2 - t1, 0.1)
        status = task_proxy.get_status()
        self.assertIsNone(status)

    def test_context_run_to_completion(self):
        """Test task stops running after completion of the task within context manager."""
        with qmi.make_task(
            "simple_task_complete",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=1.0,
            value=7,
        ) as task_proxy:
            self.assertTrue(task_proxy.is_running())
            time.sleep(2.0)
            self.assertFalse(task_proxy.is_running())
            t1 = time.monotonic()

        t2 = time.monotonic()
        self.assertLess(t2 - t1, 0.1)
        status = task_proxy.get_status()
        self.assertEqual(status, 7)

    def test_context_run_to_stopped(self):
        """Test task stops running within context manager if it is called to stop."""
        with qmi.make_task(
            "simple_task_stopped",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=2.0,
            value=8,
        ) as task_proxy:
            time.sleep(0.5)
            self.assertTrue(task_proxy.is_running())
            t1 = time.monotonic()
            task_proxy.stop()
            t2 = time.monotonic()
            self.assertLess(t2 - t1, 0.1)
            time.sleep(0.1)
            self.assertFalse(task_proxy.is_running())
            t1 = time.monotonic()

        t2 = time.monotonic()
        self.assertLess(t2 - t1, 0.1)
        status = task_proxy.get_status()
        self.assertIsNone(status)

    def test_update_settings(self):
        """Test updating task settings."""
        task_proxy = qmi.make_task(
            "simple_task_update",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=2.0,
            value=9,
        )
        recv = QMI_SignalReceiver()
        task_proxy.sig_settings_updated.subscribe(recv)
        task_proxy.start()
        time.sleep(0.5)
        old_settings = task_proxy.get_settings()
        new_settings = SimpleTestTask.Settings("hello", 101)
        task_proxy.set_settings(new_settings)
        time.sleep(0.5)
        active_settings = task_proxy.get_settings()
        task_proxy.join()
        status = task_proxy.get_status()
        settings_received = recv.get_next_signal().args[0]

        self.assertNotEqual(old_settings, active_settings)
        self.assertEqual(active_settings, new_settings)
        self.assertEqual(status, 101)
        self.assertEqual(settings_received, new_settings)

    def test_update_settings_with_context(self):
        """Test updating task settings within a context manager."""
        recv = QMI_SignalReceiver()
        with qmi.make_task(
            "simple_task_update",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=2.0,
            value=9,
        ) as task_proxy:
            task_proxy.sig_settings_updated.subscribe(recv)
            time.sleep(0.5)
            old_settings = task_proxy.get_settings()
            new_settings = SimpleTestTask.Settings("hello", 101)
            task_proxy.set_settings(new_settings)
            time.sleep(0.5)
            active_settings = task_proxy.get_settings()
            task_proxy.join()
            status = task_proxy.get_status()
            settings_received = recv.get_next_signal().args[0]

        self.assertNotEqual(old_settings, active_settings)
        self.assertEqual(active_settings, new_settings)
        self.assertEqual(status, 101)
        self.assertEqual(settings_received, new_settings)

    def test_get_pending_settings(self):
        """A more complex test to obtain pending settings for slow cycle tasks, where the new settings are not updated
        frequently."""
        event = threading.Event()
        task_proxy = qmi.make_task(
            "slow_task", SlowTestTask, duration=2.0, value=9, event=event
        )
        recv = QMI_SignalReceiver()
        task_proxy.sig_settings_updated.subscribe(recv)
        task_proxy.start()
        pre_pending_settings = task_proxy.get_pending_settings()
        new_settings = SimpleTestTask.Settings("hello", 101)
        # When setting new settings with QMI_TaskRunner.set_settings the new settings are put into a FiFo queue.
        task_proxy.set_settings(new_settings)
        # The settings won't be updated in the proxy before update_settings() has been called in a LoopTask instance.
        # This triggers obtaining the next item in the FiFo queue.
        # In the run() of SlowTestTask class, update_settings() called only after the event is set.
        pending_settings = task_proxy.get_pending_settings()
        event.set()
        time.sleep(0.4)
        self.assertIsNone(task_proxy.get_pending_settings())
        event.set()
        time.sleep(0.4)
        post_pending_settings = task_proxy.get_pending_settings()
        task_proxy.join()

        self.assertIsNone(pre_pending_settings)
        self.assertEqual(pending_settings, new_settings)
        self.assertIsNone(post_pending_settings)

    def test_get_pending_settings_with_context(self):
        """A more complex test to obtain pending settings for slow cycle tasks, where the new settings are not updated
        frequently, within a context manager."""
        event = threading.Event()
        recv = QMI_SignalReceiver()
        with qmi.make_task(
            "slow_task", SlowTestTask, duration=2.0, value=9, event=event
        ) as task_proxy:
            task_proxy.sig_settings_updated.subscribe(recv)
            pre_pending_settings = task_proxy.get_pending_settings()
            new_settings = SimpleTestTask.Settings("hello", 101)
            # When setting new settings with QMI_TaskRunner.set_settings the new settings are put into a FiFo queue.
            task_proxy.set_settings(new_settings)
            # The settings won't be updated in the proxy before update_settings() has been called in a LoopTask
            # instance. This triggers obtaining the next item in the FiFo queue.
            # In the run() of SlowTestTask class, update_settings() called only after the event is set.
            pending_settings = task_proxy.get_pending_settings()
            event.set()
            time.sleep(0.4)
            self.assertIsNone(task_proxy.get_pending_settings())
            event.set()
            time.sleep(0.4)
            post_pending_settings = task_proxy.get_pending_settings()

        self.assertIsNone(pre_pending_settings)
        self.assertEqual(pending_settings, new_settings)
        self.assertIsNone(post_pending_settings)

    def test_signals(self):
        """Test the published signal in task gets added to the queue and received correctly."""
        task_proxy = qmi.make_task(
            "simple_task_signals",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=1.0,
            value=10,
        )
        recv = QMI_SignalReceiver()
        task_proxy.my_signal.subscribe(recv)
        task_proxy.start()
        time.sleep(0.5)
        task_proxy.join()
        self.assertTrue(recv.has_signal_ready())
        self.assertEqual(recv.get_queue_length(), 1)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self._ctx_qmi_id)
        self.assertEqual(sig.publisher_name, "simple_task_signals")
        self.assertEqual(sig.signal_name, "my_signal")
        self.assertEqual(sig.args, (10,))

    def test_stop_before_start(self):
        """Test that just making a task doesn't start it, but also does not fail if `stop()` or `join()` is called."""
        task_proxy = qmi.make_task(
            "simple_task_stop",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=2.0,
            value=11,
        )
        time.sleep(0.5)
        self.assertFalse(task_proxy.is_running())
        task_proxy.stop()
        self.assertFalse(task_proxy.is_running())
        t1 = time.monotonic()
        task_proxy.join()
        t2 = time.monotonic()
        self.assertLess(t2 - t1, 0.1)
        status = task_proxy.get_status()
        self.assertIsNone(status)

    def test_receive_signals(self):
        """Test that signals are received one cycle (0.1s) after sending, and that the task stops at -1."""
        publisher_proxy = qmi.context().make_rpc_object("test_publisher", TestPublisher)
        task_proxy = qmi.make_task(
            "receiving_task",
            SignalWaitingTask,
            wait_timeout=None,
            context_id=self._ctx_qmi_id,
        )
        task_proxy.start()
        time.sleep(0.1)
        status = task_proxy.get_status()
        self.assertIsNone(status)
        publisher_proxy.send_signal(8)
        time.sleep(0.1)
        status = task_proxy.get_status()
        self.assertEqual(status, 8)
        publisher_proxy.send_signal(9)
        time.sleep(0.1)
        status = task_proxy.get_status()
        self.assertEqual(status, 9)
        publisher_proxy.send_signal(-1)
        time.sleep(0.1)
        self.assertFalse(task_proxy.is_running())
        task_proxy.join()

    def test_receive_signals_in_context(self):
        """Test that signals are received one cycle (0.1s) after sending within a context manager."""
        publisher_proxy = qmi.context().make_rpc_object("test_publisher", TestPublisher)
        with qmi.make_task(
            "receiving_task",
            SignalWaitingTask,
            wait_timeout=None,
            context_id=self._ctx_qmi_id,
        ) as task_proxy:
            self.assertTrue(task_proxy.is_running())
            time.sleep(0.1)
            status = task_proxy.get_status()
            self.assertIsNone(status)
            publisher_proxy.send_signal(8)
            time.sleep(0.1)
            status = task_proxy.get_status()
            self.assertEqual(status, 8)
            publisher_proxy.send_signal(9)
            time.sleep(0.1)
            status = task_proxy.get_status()
            self.assertEqual(status, 9)
            publisher_proxy.send_signal(-1)
            time.sleep(0.1)
            self.assertFalse(task_proxy.is_running())

    def test_timeout_while_waiting_for_signal(self):
        """Test when a wait times out, it raises an exception."""
        qmi.context().make_rpc_object("test_publisher", TestPublisher)
        task_proxy = qmi.make_task(
            "receiving_task",
            SignalWaitingTask,
            wait_timeout=1.0,
            context_id=self._ctx_qmi_id,
        )
        task_proxy.start()
        time.sleep(0.1)
        self.assertTrue(task_proxy.is_running())
        time.sleep(1.4)
        self.assertFalse(task_proxy.is_running())
        status = task_proxy.get_status()
        self.assertEqual(status, -101)  # indicates task got QMI_TimeoutException
        # task will have raised a QMI_TimeoutException while waiting for a signal
        with self.assertRaises(qmi.core.exceptions.QMI_TaskRunException):
            task_proxy.join()

    def test_timeout_while_waiting_for_signal_in_context(self):
        """Test when a wait times out, it raises an exception within a context manager."""
        qmi.context().make_rpc_object("test_publisher", TestPublisher)
        # task will have raised a QMI_TimeoutException while waiting for a signal
        with self.assertRaises(qmi.core.exceptions.QMI_TaskRunException):
            with qmi.make_task(
                "receiving_task",
                SignalWaitingTask,
                wait_timeout=1.0,
                context_id=self._ctx_qmi_id,
            ) as task_proxy:
                time.sleep(0.1)
                self.assertTrue(task_proxy.is_running())
                time.sleep(1.4)
                self.assertFalse(task_proxy.is_running())
                status = task_proxy.get_status()
                self.assertEqual(status, -101)  # indicates task got QMI_TimeoutException

    def test_stop_while_waiting_for_signal(self):
        qmi.context().make_rpc_object("test_publisher", TestPublisher)
        task_proxy = qmi.make_task(
            "receiving_task",
            SignalWaitingTask,
            wait_timeout=30.0,
            context_id=self._ctx_qmi_id,
        )
        task_proxy.start()
        time.sleep(0.1)
        self.assertTrue(task_proxy.is_running())
        task_proxy.stop()
        time.sleep(0.1)
        self.assertFalse(task_proxy.is_running())
        status = task_proxy.get_status()
        self.assertEqual(status, -100)  # indicates task got QMI_TaskStopException
        task_proxy.join()

    def test_run_to_loop_finish(self):
        """Test and assert that loop runs normally from start to finish if there are no glitches."""
        # Arrange
        status_signals_received = []
        settings_signals_received = []
        nr_of_loops = 3
        increase_loop = False
        initial_status_value = -1
        loop_period = 0.2
        policy = QMI_LoopTaskMissedLoopPolicy.IMMEDIATE
        status_signals_expected = list(range(initial_status_value, nr_of_loops, 1)) + [1]
        settings_signals_expected = [1] + list(range(1, 4))

        task_proxy = qmi.make_task(
            "loop_task_finish",
            LoopTestTask,
            increase_loop=increase_loop,
            nr_of_loops=nr_of_loops,
            status_value=initial_status_value,
            loop_period=loop_period,
            policy=policy,
        )
        receiver = QMI_SignalReceiver()
        publisher_proxy = qmi.get_task(f"{self._ctx_qmi_id}.loop_task_finish")
        publisher_proxy.sig_settings_updated.subscribe(receiver)

        # Act
        task_proxy.start()
        # Test that prepare has done its job
        status_signals_received.append(task_proxy.get_status().value)
        settings_signals_received.append(receiver.get_next_signal(timeout=loop_period).args[-1])
        # LoopTestTask does 3x 1 second loops --> should be finished after 3 seconds
        for n in range(nr_of_loops):
            while (status := task_proxy.get_status().value) == status_signals_received[-1]:
                pass  # Do as fast as possible

            status_signals_received.append(status)
            if not receiver.has_signal_ready():
                time.sleep(loop_period - (time.monotonic() % loop_period))  # synchronize

            # Test that the status changes at the end of each loop, after receiver signal increments
            settings_signals_received.append(receiver.get_next_signal(timeout=loop_period).args[-1])

        # Test that finalize_loop sets status back to 1
        time.sleep(loop_period)
        status_signals_received.append(task_proxy.get_status().value)
        task_proxy.join()  # Give time to finalize
        # Assert
        self.assertListEqual(status_signals_expected, status_signals_received)
        self.assertListEqual(settings_signals_expected, settings_signals_received)
        self.assertFalse(task_proxy.is_running())

    def test_run_to_loop_finish_in_context(self):
        """Test and assert that loop runs normally within context manager if there are no glitches."""
        # Arrange
        status_signals_received = []
        settings_signals_received = []
        nr_of_loops = 3
        increase_loop = False
        initial_status_value = -1
        loop_period = 0.2
        policy = QMI_LoopTaskMissedLoopPolicy.IMMEDIATE
        status_signals_expected = list(range(initial_status_value + 1, nr_of_loops, 1)) + [1]
        settings_signals_expected = list(range(1, 4))
        settings_receiver = QMI_SignalReceiver()
        status_receiver = QMI_SignalReceiver()

        with qmi.make_task(
            "loop_task_finish_with",
            LoopTestTask,
            increase_loop=increase_loop,
            nr_of_loops=nr_of_loops,
            status_value=initial_status_value,
            loop_period=loop_period,
            policy=policy,
        ) as task_proxy:
            # Act
            publisher_proxy = qmi.get_task(f"{self._ctx_qmi_id}.loop_task_finish_with")
            publisher_proxy.sig_status_updated.subscribe(status_receiver)
            publisher_proxy.sig_settings_updated.subscribe(settings_receiver)
            # Test that prepare has done its job
            status_init = task_proxy.get_status().value
            # LoopTestTask does 3x 1 second loops --> should be finished after 3 seconds
            for n in range(nr_of_loops):
                # Test that the status changes at the end of each loop, after receiver signal increments
                while not status_receiver.has_signal_ready():
                    time.sleep(loop_period - (time.monotonic() % loop_period))  # Synchronize

                status = status_receiver.get_next_signal(timeout=loop_period).args[-1]
                if status == status_init:
                    time.sleep(loop_period - (time.monotonic() % loop_period))  # Synchronize
                    status = status_receiver.get_next_signal(timeout=loop_period).args[-1]

                while not settings_receiver.has_signal_ready():
                    time.sleep(loop_period - (time.monotonic() % loop_period))

                setting = settings_receiver.get_next_signal(timeout=loop_period).args[-1]
                if len(settings_signals_received) and setting == settings_signals_received[-1]:
                    setting = settings_receiver.get_next_signal(timeout=loop_period).args[-1]

                status_signals_received.append(status)
                settings_signals_received.append(setting)

            # Test that finalize_loop sets status back to 1
            time.sleep(loop_period)
            status_signals_received.append(task_proxy.get_status().value)

        # Assert
        self.assertEqual(initial_status_value, status_init)
        self.assertListEqual(status_signals_expected, status_signals_received)
        self.assertListEqual(settings_signals_expected, settings_signals_received)
        self.assertFalse(task_proxy.is_running())

    def test_my_signal(self):
        """See that my_signal also gets updated and sent in each loop with increasing step"""
        # Arrange
        increase_loop = True  # +1 extra status value / round
        nr_of_loops = 5
        initial_status_value = 5
        loop_period = 0.1
        policy = QMI_LoopTaskMissedLoopPolicy.IMMEDIATE
        my_signals_expected = list(range(
            initial_status_value + 2, initial_status_value + nr_of_loops * 2, 3)
        )  # my_signal sums up (increasing) status and settings values, starting after 1st round.
        my_signals_received = []

        task_proxy = qmi.make_task(
            "loop_task_my_signal",
            LoopTestTask,
            increase_loop=increase_loop,
            nr_of_loops=nr_of_loops,
            status_value=initial_status_value,
            loop_period=loop_period,
            policy=policy,
        )
        receiver = QMI_SignalReceiver()
        publisher_proxy = qmi.get_task(f"{self._ctx_qmi_id}.loop_task_my_signal")
        publisher_proxy.my_signal.subscribe(receiver)
        LoopTestTask.Settings("Standard setting for LoopTestTask", 1)

        # Act
        task_proxy.start()
        for n in range(len(my_signals_expected)):
            # Test that the my_signal is sent at each loop with increasing value
            while not receiver.has_signal_ready():
                time.sleep(loop_period - (time.monotonic() % loop_period))  # synchronize

            my_signals_received.append(receiver.get_next_signal(timeout=loop_period).args[-1])

        # Assert
        self.assertListEqual(my_signals_expected, my_signals_received)

    def test_my_signal_timeout(self):
        """See that my_signal raises timeout if trying to wait for more signals than the loops"""
        # Arrange
        increase_loop = True
        nr_of_loops = 2
        initial_status_value = 0
        loop_period = 0.1
        policy = QMI_LoopTaskMissedLoopPolicy.IMMEDIATE

        task_proxy = qmi.make_task(
            "loop_task",
            LoopTestTask,
            increase_loop=increase_loop,
            nr_of_loops=nr_of_loops,
            status_value=initial_status_value,
            loop_period=loop_period,
            policy=policy,
        )
        receiver = QMI_SignalReceiver()
        publisher_proxy = qmi.get_task(f"{self._ctx_qmi_id}.loop_task")
        publisher_proxy.my_signal.subscribe(receiver)
        LoopTestTask.Settings("Standard setting for LoopTestTask", 1)

        # Act
        task_proxy.start()

        # Assert
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            for n in range(3):
                # Test that the my_signal is sent at each loop with increasing loop jumps
                _ = receiver.get_next_signal(timeout=2 * loop_period).args[-1]

    def test_loop_timeout_policy_IMMEDIATE(self):
        """See that all loop iterations are obtained even if with some delay"""
        # Arrange
        nr_of_loops = 5
        increase_loop = False
        initial_status_value = -1
        loop_period = 0.1
        policy = QMI_LoopTaskMissedLoopPolicy.IMMEDIATE

        status_signals_expected = list(range(
            initial_status_value, initial_status_value + nr_of_loops + 1, 1)
        )
        settings_signals_expected = [1] + list(range(1, nr_of_loops + 1))
        status_got = []
        settings_got = []
        status_signals_received = []
        settings_signals_received = []

        # The `error=IMMEDIATE` in the task object creation makes each loop be 1.1 times longer than loop period.
        task_proxy = qmi.make_task(
            "loop_task_immediate",
            LoopTestTask,
            increase_loop=increase_loop,
            nr_of_loops=nr_of_loops,
            status_value=initial_status_value,
            loop_period=loop_period,
            policy=policy,
            error="IMMEDIATE",
        )
        receiver_sett = QMI_SignalReceiver()
        receiver_stat = QMI_SignalReceiver()
        publisher_proxy = qmi.get_task(f"{self._ctx_qmi_id}.loop_task_immediate")
        publisher_proxy.sig_settings_updated.subscribe(receiver_sett)
        publisher_proxy.sig_status_updated.subscribe(receiver_stat)

        # Act
        # Get initial values before start
        settings_got.append(task_proxy.get_settings().value)
        status_got.append(task_proxy.get_status().value)
        start = time.time()
        task_proxy.start()
        # Test that prepare has done its job
        settings_signals_received.append(
            receiver_sett.get_next_signal(timeout=2 * loop_period).args[-1]
        )
        status_signals_received.append(
            receiver_stat.get_next_signal(timeout=2 * loop_period).args[-1]
        )
        # LoopTestTask does 1.1x loop_period --> should be finished after nr_of_loops*loop_period
        for n in range(nr_of_loops):
            # Test that the status changes at the end of each loop, after receiver signal increments
            settings_got.append(task_proxy.get_settings().value)
            while task_proxy.get_status().value == status_got[-1]:
                time.sleep(0.1 * loop_period)

            status_got.append(task_proxy.get_status().value)
            settings_signals_received.append(
                receiver_sett.get_next_signal(timeout=2 * loop_period).args[-1]
            )
            status_signals_received.append(
                receiver_stat.get_next_signal(timeout=loop_period).args[-1]
            )

        time.sleep(loop_period * 2)  # Need to sleep once more for waiting the last loop `sleep`.
        # Test that finalize_loop sets settings and status back to 1
        settings_got.append(task_proxy.get_settings().value)
        status_got.append(task_proxy.get_status().value)
        end = time.time()
        # Assert
        self.assertFalse(task_proxy.is_running())
        self.assertGreater(end - start, 3 * loop_period)
        self.assertListEqual(settings_signals_received, settings_got[:-1])
        self.assertListEqual(status_signals_expected, status_got[:-1])
        self.assertEqual(1, settings_got[-1])  # loop_finalize should set the last value as 1
        self.assertEqual(1, status_got[-1])  # loop_finalize should set the last value as 1
        self.assertListEqual(status_signals_expected, status_signals_received)
        self.assertListEqual(settings_signals_expected, settings_signals_received)

    def test_loop_timeout_policy_SKIP(self):
        """See that setting policy to skip causes to skip loops"""
        # Arrange
        status_signals_received = []
        settings_signals_received = []
        increase_loop = False
        nr_of_loops = 5
        initial_status_value = -1
        loop_period = 0.1
        policy = QMI_LoopTaskMissedLoopPolicy.SKIP

        status_signals_expected = list(range(initial_status_value, round(nr_of_loops / 2 + 1e-15)))
        settings_signals_expected = [1] + [1, 3, 5]  # default value + skips loops 2 & 4

        task_proxy = qmi.make_task(
            "loop_task",
            LoopTestTask,
            increase_loop=increase_loop,
            nr_of_loops=nr_of_loops,
            status_value=initial_status_value,
            loop_period=loop_period,
            policy=policy,
            error="SKIP",
        )
        receiver_sett = QMI_SignalReceiver()
        receiver_stat = QMI_SignalReceiver()
        publisher_proxy = qmi.get_task(f"{self._ctx_qmi_id}.loop_task")
        publisher_proxy.sig_settings_updated.subscribe(receiver_sett)
        publisher_proxy.sig_status_updated.subscribe(receiver_stat)

        # Act
        status_got = task_proxy.get_status().value
        task_proxy.start()
        # Test that prepare has done its job
        settings_signals_received.append(
            receiver_sett.get_next_signal(timeout=loop_period).args[-1]
        )
        status_signals_received.append(
            receiver_stat.get_next_signal(timeout=loop_period).args[-1]
        )
        # LoopTestTask does 1.1x loop_period --> should be finished after half, rounded up, nr_of_loops has been done.
        for n in range(round(nr_of_loops / 2 + 1e-15)):  # + 1e-15 to cheat on "banker's rounding" to "always" round up
            # Test that the status changes at the end of each loop, after receiver signal increments
            while task_proxy.get_status().value == status_got:
                time.sleep(0.1 * loop_period)

            status_got = task_proxy.get_status().value
            status_signals_received.append(
                receiver_stat.get_next_signal(timeout=2 * loop_period).args[-1]
            )
            settings_signals_received.append(
                receiver_sett.get_next_signal(timeout=2 * loop_period).args[-1]
            )

        time.sleep(loop_period * 2)  # Make sure the loop has time to finalize
        # Assert
        # Test that finalize_loop sets status back to 1
        self.assertFalse(task_proxy.is_running())
        self.assertEqual(1, task_proxy.get_settings().value)
        self.assertEqual(1, task_proxy.get_status().value)
        self.assertListEqual(status_signals_expected, status_signals_received)
        self.assertListEqual(settings_signals_expected, settings_signals_received)

    def test_loop_timeout_policy_TERMINATE(self):
        """See that setting policy to terminate stops the task"""
        # Arrange
        status_signals_received = []
        settings_signals_received = []
        increase_loop = False
        nr_of_loops = 5
        initial_status_value = 3
        loop_period = 0.1
        policy = QMI_LoopTaskMissedLoopPolicy.TERMINATE

        status_signals_expected = list(range(
            initial_status_value, initial_status_value + 2, 1)
        ) + [1]  # Expected to skip a sample on 2nd step and terminate, last value defaults back to 1
        settings_signals_expected = [1, 1]  # initial value, 1st loop

        task_proxy = qmi.make_task(
            "loop_task_terminate",
            LoopTestTask,
            increase_loop=increase_loop,
            nr_of_loops=nr_of_loops,
            status_value=initial_status_value,
            loop_period=loop_period,
            policy=policy,
            error="TERMINATE",
        )
        receiver = QMI_SignalReceiver()
        publisher_proxy = qmi.get_task(f"{self._ctx_qmi_id}.loop_task_terminate")
        publisher_proxy.sig_settings_updated.subscribe(receiver)

        # Act
        # Test that initial status value is set as -1
        status_signals_received.append(task_proxy.get_status().value)
        task_proxy.start()
        # LoopTestTask does 5x 1 second loops, but we should miss every second loop --> should be finished after 2 loops
        for n in range(2):
            settings_signals_received.append(
                receiver.get_next_signal(timeout=2 * loop_period).args[-1]
            )
            # Test that the status changes at the end of each loop, after receiver signal increments
            while task_proxy.get_status().value == status_signals_received[-1]:
                time.sleep(0.1 * loop_period)

            status_signals_received.append(task_proxy.get_status().value)

        # 3rd loop should raise the timeout exception
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            settings_signals_received.append(
                receiver.get_next_signal(timeout=loop_period).args[-1]
            )

        time.sleep(loop_period)  # The loop should terminate with the timeout
        # Assert
        self.assertFalse(task_proxy.is_running())
        self.assertListEqual(status_signals_expected, status_signals_received)
        self.assertListEqual(settings_signals_expected, settings_signals_received)

    def test_get_task_class_name(self):
        task_proxy = qmi.make_task(
            "simple_task_name",
            SimpleTestTask,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=1.0,
            value=7,
        )
        task_class_name = task_proxy.get_task_class_name()
        expect_name = __name__ + ".SimpleTestTask"
        self.assertEqual(task_class_name, expect_name)

    def test_custom_task_runner(self):
        task_proxy = qmi.make_task(
            "simple_task_runner",
            SimpleTestTask,
            task_runner=CustomTaskRunner,
            raise_exception_in_init=False,
            raise_exception_in_run=False,
            duration=1.0,
            value=7,
        )
        actual = task_proxy.get_custom_taskrunner_attr()
        self.assertEqual(actual, sentinel.custom_taskrunner_attr)


if __name__ == "__main__":
    unittest.main()



"""
Tenma (Newark Electronics) power supply units.

The qmi.instruments.tenma package provides support for:
- Series 72 power supply unit groups
  - 72-2535, 72-2540, 72-2545, 72-2550, 72-2925, 72-2930, 72-2935, 72-2940 & 72-10480
  - 72-13350 & 72-13360
"""
from qmi.instruments.tenma.psu_72 import *



"""
Newport Corporation, MKS and New Focus; motion controllers, steppers, power meter, tunable laser controller.

The qmi.instruments.newport package provides support for:
- Agilis AG-UC8 piezo stepper controller
- Motion controllers (including, but not limited to):
  - SMC100CC, with linear actuator:
    - CMA25CCL
  - CONEX-CC, with linear actuators:
    - TRA12CC
    - TRB6CC
- 843-R optical power meter
- TLB-6700 series tunable laser controllers.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.newport.ag_uc8 import Newport_AG_UC8 as Newport_AgilisUc8
from qmi.instruments.newport.ag_uc8 import AxisStatus
from qmi.instruments.newport.smc_100cc import Newport_SMC100CC as Newport_Smc100Cc, ControlLoopState
from qmi.instruments.newport.smc_100pp import Newport_SMC100PP as Newport_Smc100Pp
from qmi.instruments.newport.conex_cc import Newport_ConexCC as Newport_ConexCc
from qmi.instruments.newport.newport_843r import Newport_843R
from qmi.instruments.newport.tlb670x import NewFocus_TLB670X as NewFocus_Tlb670x
